"""Contains object representation of possible variable transformations."""

import pandas as pd
import numpy as np


# This is a class to package together the logic of how to transform
# with how to display / print the transformation

class Transformation():
    """A Trasformation object holds the actual function to calculate
    transformations on data as well as some helper information for
    printing and visualizing.
    """
    
    def __init__(self, func, pattern, name, inverse=None):
        """Creates a Transformation object.

        Arguments:
            func - A function to be applied to a column of data in a DataFrame.
            pattern - A str holding a template for printing.
            name - A str describing the transformation.
            inverse - An optional function that will undo the func operation.
        """
        self.func = func
        self.pattern = pattern
        self.inverse = inverse
        self.name = name
        
    def __str__(self):
        """Returns the given pattern for debugging. """
        return self.pattern

    def __eq__(self, other):
        """Check if two objects are equivalent. 

        Arguments:
            other - An object.

        Returns:
            A boolean, True if other is a Transformation and equivalent, False
            if else.
        """
        if isinstance(other, Transformation):
            return self.pattern == other.pattern and self.name == other.name
        else:
            return False

    def __hash__(self):
        """Hash the transformation for the purposes of storing in sets and
        dictionaries.

        Returns:
            A real value representing the hash of the Transformation.
        """

        return hash((self.pattern, self.name))
        
    def compose(self, inner):
        """Applies specific data to the pattern for printing. 
        
        Arguments:
            inner - An object to be injested by str.format()
        
        Returns:
            A processed and evaluated str representing the Trasnformation.
        """
        return self.pattern.format(inner)
        
    def transform(self, values, training=True):
        """Apply the function to the data.

        Arguments:
            values - A Series object that is the data to trasnform.
            training - A flag to indicate is this transformation is during
                training or not. Default is True.

        Returns:
            A trasnformed Series. 
        """
        return self.func(values)
    
    def copy(self):
        """Returns a deep copy of the Transformation."""
        return Transformation(self.func, self.pattern, self.name, self.inverse)

    def invert(self, data):
        """If available, invert the data.

        Arguments:
            data - A Series object holding the data to invert.

        Returns:
            A Series object that has the inverted data.
        """
        if self.inverse is None:
            raise Exception("Inverse not defined for " + self.name + " transformation.")

        return self.inverse(data)

class Center(Transformation):
    """A specific type of Trasnformation for centering data so that it has a
    mean of 0."""

    def __init__(self):
        """Create a Center object."""
        self.pattern = "{0}-E({0})"
        self.past_mean = 0
        self.name = "Center"
        
    def transform(self, values, training=True):
        if training:
            self.past_mean = values.mean()
        
        return values - self.past_mean
    
    def copy(self):
        ret_val = Center()
        ret_val.past_mean = self.past_mean
        return ret_val

    def invert(self, data):
        return data + self.past_mean

class Standardize(Transformation):
    """A specific type of Transformation that standardizes the data so that it
    has a mean of 0 and standard deviation of 1."""
    
    def __init__(self):
        """Create a Standardize object. """
        self.pattern = "({0}-E({0}))/Std({0})"
        self.past_mean = 0
        self.past_std = 1
        self.name = "Standardize"
        
    def transform(self, values, training = True):
        if training:
            self.past_mean = values.mean()
            self.past_std = values.std()
            
        return (values - self.past_mean) / self.past_std    
    
    def copy(self):
        ret_val = Standardize()
        ret_val.past_mean, ret_val.past_std = self.past_mean, self.past_std
        return ret_val

    def invert(self, data):
        return (data * self.past_std) + self.past_mean
    
# Aliases for common Transformations
Sin = lambda i: Transformation(np.sin, "sin({})", "Sine")
Cos = lambda i: Transformation(np.cos, "cos({})", "Cosine")
Log = lambda i: Transformation(np.log, "log({})", "Natural Log", np.exp)
Log10 = lambda i: Transformation(np.log10, "log({})", "Log Base 10", lambda x: 10 * x)
Exp = lambda i: Transformation(np.exp, "exp({})", "Exponential", np.log)
Std = lambda i: Standardize()
Cen = lambda i: Center()
Identity = lambda i: Transformation(lambda x: x, "{}", "Identity", lambda x: x)
Increment = lambda i: Transformation(lambda x: x + i, "{}+"+str(i) if i >= 0 else "{}-"+str(-i), "Increment", lambda x: x - i)
Multiply = lambda i: Transformation(lambda x: x * i, str(i) + "*{}", "Multiply", lambda x: x * (1/i))
Power = lambda i: Transformation(lambda x: x ** i, "{}^" + str(i), "Power", lambda x: x ** (1/i) if i % 2 == 1 else x.clip(0, None) ** (1/i))

_default_transformations = {
    "sin" : Sin,
    "cos" : Cos,
    "log" : Log,
    "log10" : Log10,
    "exp" : Exp,
    "standardize" : Std,
    "center": Cen,
    "identity": Identity
}


"""Describes various linear models supported by SALMON."""

import numpy as np

import scipy.stats as stats
from scipy.linalg import solve_triangular, cho_solve

import pandas as pd
from pandas.plotting import scatter_matrix

import matplotlib.pyplot as plt
import warnings

from itertools import product
from collections import OrderedDict
from ordered_set import OrderedSet

from .expression import Expression, Combination, Identity, Constant, LightDataFrame, Quantitative

plt.style.use('ggplot')


def _float_format(x):
    abs_x = abs(x)
    if abs_x >= 1e4:
        rep = "{:.3e}".format(x)
    elif abs_x >= 1e0:
        rep = "{:."
        rep += str(3 - int(np.floor(np.log10(abs_x))))
        rep += "f}"
        rep = rep.format(x)
    elif abs_x >= 1e-3:
        rep = "{:.4f}".format(x)
    elif abs_x >= 1e-9:
        rep = "{:.3e}".format(x)
    elif abs_x >= 1e-99:
        rep = "{:.1e}".format(x)
    else:
        rep = "{:.0e}".format(x)
    rep = rep.replace("e-0", "e-").replace("e+0", "e+")
    rep = rep.replace("0e+0", "0.000")
    return rep


pd.set_option("display.float_format", _float_format)


def _confint(estimates, standard_errors, df, crit_prob):
    crit_value = stats.t.ppf(crit_prob, df)
    ci_widths = crit_value * standard_errors
    return estimates - ci_widths, estimates + ci_widths


def qr_solve(Q, R, y):
    """Solve least squares X / y, given QR decomposition of X"""
    _, p = R.shape
    if p:
        return solve_triangular(R, Q.T @ y, check_finite=False)
    else:
        return np.empty(shape=0)


def cho_inv(R):
    """Calculate inverse of X.T @ X, given Cholesky decomposition R.T @ R"""
    _, p = R.shape
    if p:
        return cho_solve((R, False), np.identity(p), check_finite=False)
    else:
        return np.empty(shape=(0, 0))


class Model:
    """A general Model class that both Linear models and (in the future)
    General Linear models stem from."""

    def __init__(self):
        """Create a Model object (only possible through inheritance)."""
        raise NotImplementedError()

    def fit(self, data):
        """Fit a model to given data.

        Arguments:
            data - A DataFrame with column names matching specified terms
                within the Model's explanatory and response Expression
                objects.

        Returns:
            A DataFrame with relevant statistics of fitted Model (coefficients,
            t statistics, p-values, etc.).
        """
        raise NotImplementedError()

    def predict(self, data):
        """Predict response values for a given set of data.

        Arguments:
            data - A DataFrame with column names matching specified terms
                within the Model's explanatory Expression object.

        Returns:
            A Series of the predicted values.
        """
        raise NotImplementedError()

    def plot_matrix(self, **kwargs):
        """Produce a matrix of pairwise scatter plots of the data it was fit
        on. The diagonal of the matrix will feature histograms instead of
        scatter plots.

        Arguments:
            kwargs - One or more named parameters that will be ingested by
                Pandas' scatter_matrix plotting function.

        Returns:
            A matplotlib plot object containing the matrix of scatter plots.
        """
        df = pd.concat([self.X_train_, self.y_train_], axis=1)
        scatter_matrix(df, **kwargs)


class LinearModel(Model):
    """A specific Model that assumes the response variable is linearly related
    to the explanatory variables."""

    def __init__(self, explanatory, response, intercept=True):
        """Create a LinearModel object.

        An intercept is included in the model by default. To fit a model
        without an intercept term, either set intercept=False or subtract '1'
        from the explanatory Expression.

        Arguments:
            explanatory - Explanatory variables used to condition the model on.
                Represented either by an `Expression` that is either a single 
                term or a `Combination` of terms, or data in the form of a 
                pandas `DataFrame` or numpy 1-D / 2-D `ndarray`. 
            response - Response variables used as the target for modeling. 
                Represented either by an `Expression` object, or data in the
                form of a pandas `Series` or numpy 1-D `ndarray` object.
            intercept - A boolean indicating whether an intercept should be
                included (True) or not (False).
        """
        # may be overwritten if `explanatory` and `response` are given as 
        # data structures rather than expressions
        self.training_data = None  
        
        if explanatory is None:
            explanatory = 0

        if isinstance(explanatory, (int, float)):
            explanatory = Constant(explanatory)
        
        if (isinstance(explanatory, (pd.DataFrame, np.ndarray)) or 
                isinstance(response, (pd.Series, np.ndarray))):

            assert(isinstance(explanatory, (pd.DataFrame, np.ndarray)) and 
                isinstance(response, (pd.Series, np.ndarray)))

            if isinstance(explanatory, np.ndarray):
                if len(explanatory.shape) == 1:
                    explanatory = explanatory.reshape((len(explanatory), 1))
                else:
                    assert(len(explanatory.shape) == 2)

                expl_data = explanatory
                expl_cols = ["x{}".format(i+1) for i in range(expl_data.shape[1])]
            else:  # isinstance(explanatory, pd.DataFrame)
                expl_data = explanatory.values
                expl_cols = list(explanatory.columns)

            if isinstance(response, np.ndarray):
                if len(response.shape) == 1:
                    response = response.reshape((len(response), 1))
                else:
                    assert(response.shape[1] == 1)
                    assert(len(response.shape) == 2)

                resp_data = response
                resp_cols = ["y"]
            else:  # isinstance(response, pd.Series)
                resp_data = response.values.reshape((len(response), 1))
                if response.name is None:
                    resp_cols = ["y"]
                else:
                    resp_cols = [response.name]
            
            self.X_train_ = LightDataFrame(expl_data, columns=expl_cols)
            self.y_train_ = LightDataFrame(resp_data, columns=resp_cols)[:, 0]
            self.training_data = LightDataFrame(
                np.hstack([expl_data, resp_data]),
                columns=expl_cols+resp_cols,
            )
            
            # Bypass construction because we know explicitly that the terms will not collide
            explanatory = Combination(terms=[])
            explanatory.terms = OrderedSet(Quantitative(c) for c in expl_cols) 
            if intercept:
                explanatory.terms.add(Constant(1))
            response = Quantitative(resp_cols[0])
        else:
            if intercept:
                explanatory = explanatory + 1

        if explanatory == Constant(0):
            raise Exception(
                "Must have at least one predictor in explanatory "
                "expression and/or intercept enabled for a valid model."
            )
        self.given_ex = explanatory

        constant = self.given_ex.reduce()['Constant']
        self.intercept = constant is not None
        if self.intercept:
            if isinstance(self.given_ex, Combination) and len(self.given_ex.terms) > 1:
                # This bypasses some potentially expensive checks 
                self.given_ex.terms.remove(Constant(constant))
            else:
                # This was done to easily check all options for indicating
                # a wanted intercept
                self.given_ex = self.given_ex - constant

        # This will collapse any combination of variables into a single column
        self.given_re = Identity(response)
        self.ex = None
        self.re = None


        self.categorical_levels = dict()

    def __str__(self):
        """Convert a LinearModel to a str format for printing."""
        if self.intercept:
            return str(self.given_re) + " ~ " + str(1 + self.given_ex)
        else:
            return str(self.given_re) + " ~ " + str(self.given_ex)

    def fit(self, X=None, y=None):
        """Fit a LinearModel to data.

        Data can either be provided as a single DataFrame X that contains both
        the explanatory and response variables, or in separate data
        structures, one containing the explanatory variables and the other
        containing the response variable. The latter is implemented so that
        LinearModel can be used as scikit-learn Estimator.

        It is fine to have extra columns in the DataFrame that are not used by
        the model---they will simply be ignored.

        Arugments:
            X - An optional DataFrame containing all of the explanatory
                variables in the model and possibly the response variable 
                too. If not given, the model is assumed to have been
                instantiated with the data in the constructor.
            y - An optional Series that contains the response variable.

        Returns:
            A DataFrame containing relevant statistics of fitted Model (e.g.,
            coefficients, p-values).
        """
        if X is None:
            X = self.X_train_
            y = self.y_train_
            self.re = self.given_re
            self.ex = self.given_ex
        else:
            if y is None:
                y = X

            # Initialize the categorical levels
            self.categorical_levels = dict()
            self.training_data = X

            # Replace all Var's with either Q's or C's
            self.re = self.given_re.copy().interpret(y)
            self.ex = self.given_ex.copy().interpret(X)

            # Construct X matrix
            X = self.ex.evaluate(X)
            self.X_train_ = X
            # Construct y vector
            y = self.re.evaluate(y)[:, 0]
            self.y_train_ = y

        # Perform linear algebra to find coefficients
        coef_, cols = self._fit(X, y)

        # Now collect other desired statistics
        # Get standard errors (diagonal of the covariance matrix)
        se_coef_ = np.sqrt(np.diagonal(self.cov_))

        # Get inference for coefficients
        self.t_ = coef_ / se_coef_
        self.p_ = 2 * stats.t.cdf(-abs(self.t_), self.rdf)
        lower_bound, upper_bound = _confint(coef_, se_coef_, self.rdf, .975)

        # Create output table
        table = pd.DataFrame(OrderedDict((
            ("Coefficient", coef_), ("SE", se_coef_),
            ("t", self.t_), ("p", self.p_),
            ("2.5%", lower_bound), ("97.5%", upper_bound)
        )), index=cols)

        self.coef_ = table["Coefficient"]
        self.se_coef_ = table["SE"]

        return table

    def _fit(self, X, y):
        if np.isnan(X).any() or np.isnan(y).any():
            warnings.warn("NaN's detected in data. Estimated coefficients may not be accurate.")

        # Get dimensions
        self.n, self.p = X.shape

        # Center if there is an intercept
        if self.intercept:
            X_offsets = X.mean(axis=0)
            y_offset = y.mean()
            X = X - X_offsets[np.newaxis, :]
        else:
            X_offsets = 0
            y_offset = 0

        # Get coefficients using QR decomposition
        q, r = np.linalg.qr(X)
        coef_ = qr_solve(q, r, y - y_offset)
        cols = X.columns.copy()  # column names

        # Get fitted values and residuals
        self.fitted_ = y_offset + np.dot(X, coef_)
        self.residuals_ = y - self.fitted_

        # Get residual variance
        self.rdf = self.n - self.p - (1 if self.intercept else 0)
        self.resid_var_ = (self.residuals_ ** 2).sum() / self.rdf

        # Get covariance matrix between coefficients
        self.cov_ = self.resid_var_ * cho_inv(r)

        # Update coefficients and covariance matrix with intercept
        # (if applicable)
        if self.intercept:
            cols.append("Intercept")
            coef_ = np.append(coef_, y_offset - (X_offsets * coef_).sum())
            cov_coef_intercept = -1*np.dot(self.cov_, X_offsets)

            var_intercept = self.resid_var_ / self.n
            var_intercept -= (X_offsets * cov_coef_intercept).sum()

            self.cov_ = np.block([
                [self.cov_, cov_coef_intercept[:, np.newaxis]],
                [cov_coef_intercept[np.newaxis, :], var_intercept]
            ])

        # TODO: Return the covariance rather than save it
        return coef_, cols

    def likelihood(self, data=None):
        """Calculate likelihood for a fitted model on either original data or
        new data."""
        return np.exp(self.log_likelihood(data))

    def log_likelihood(self, data=None):
        """Calculate a numerically stable log_likelihood for a fitted model
        on either original data or new data."""

        if data is None:
            residuals = self.residuals_
        else:
            y = self.re.evaluate(data)
            y_hat = self.predict(
                data,
                for_plot=False,
                confidence_interval=False,
                prediction_interval=False,
            )
            residuals = y[:, 0] - y_hat.iloc[:, 0]

        n = len(residuals)

        return (-n / 2 * (np.log(2 * np.pi) + np.log(self.resid_var_)) -
                (1 / (2 * self.resid_var_)) * (residuals ** 2).sum())

    def confidence_intervals(self, alpha=None, conf=None):
        """Calculate confidence intervals for the coefficients.

        This function assumes that Model.fit() has already been called.

        Arguments:
            alpha - A float between 0.0 and 1.0 representing the non-coverage
                probability of the confidence interval. In other words, the
                confidence level is 1 - alpha / 2.
            conf - A float between 0.0 and 1.0 representing the confidence
                level. Only one of alpha or conf needs to be specified. If
                neither are specified, a default value of conf=0.95 will be
                used.

        Returns:
            A DataFrame containing the appropriate confidence intervals for
            all the coefficients.
        """
        if alpha is None:
            if conf is None:
                conf = 0.95
            alpha = 1 - conf

        crit_prob = 1 - (alpha / 2)

        lower_bound, upper_bound = _confint(self.coef_, self.se_coef_,
                                            self.rdf, crit_prob)

        return pd.DataFrame({
            "%.1f%%" % (100 * (1 - crit_prob)): lower_bound,
            "%.1f%%" % (100 * crit_prob): upper_bound
        }, index=self.coef_.index)

    def predict(
        self,
        data,
        for_plot=False,
        confidence_interval=False,
        prediction_interval=False,
    ):
        """Predict response values from a fitted Model.

        Arguments:
            data - A DataFrame containing the values of the explanatory
                variables, for which predictions are desired.
            for_plot - A boolean indicating if these predictions are computed
                for the purposes of plotting.
            confidence_interval - If a confidence interval for the mean
                response is desired, this is a float between 0.0 and 1.0
                indicating the confidence level to use.
            prediction_interval - If a prediction interval is desired, this is
                a float between 0.0 and 1.0 indicating the confidence level to
                use.

        Returns:
            A DataFrame containing the predictions and/or intervals.
        """
        # Construct the X matrix
        if isinstance(data, (LightDataFrame, pd.DataFrame)):
            X = self.ex.evaluate(data, fit=False)
        else:
            assert(data.shape[-1] == self.X_train_.shape[-1])
            X = data

        if self.intercept:
            n, _ = X.shape
            X = np.hstack((X, np.ones((n, 1))))
        y_vals = np.dot(X, self.coef_)

        if isinstance(data, pd.DataFrame):
            index = data.index
        else:
            index = None
        
        predictions = pd.DataFrame(
            {"Predicted " + str(self.re): y_vals},
            index=index,
        )

        if confidence_interval or prediction_interval:
            if confidence_interval:
                alpha = confidence_interval
                widths = self._confidence_interval_width(
                    X,
                    confidence_interval,
                )
            else:
                alpha = prediction_interval
                widths = self._prediction_interval_width(
                    X,
                    prediction_interval,
                )

            crit_prob = 1 - (alpha / 2)

            lower = y_vals - widths
            upper = y_vals + widths

            predictions[str(round(1 - crit_prob, 5) * 100) + "%"] = lower
            predictions[str(round(crit_prob, 5) * 100) + "%"] = upper

        return predictions

    def get_sse(self):
        """Get the SSE of a fitted model."""
        sse = ((self.y_train_ - self.fitted_) ** 2).sum()
        return sse

    def get_ssr(self):
        """Get the SSR of a fitted model."""
        ssr = self.get_sst() - self.get_sse()
        return ssr

    def get_sst(self):
        """Get the SST of a fitted model."""
        sst = ((self.y_train_ - self.y_train_.mean()) ** 2).sum()
        return sst

    def r_squared(self, X=None, y=None, adjusted=False, **kwargs):
        """Calculate the (adjusted) R^2 value of the model.
        This can be used as a metric within the sklearn ecosystem.

        Arguments:
            X - An optional DataFrame of the explanatory data to be used for
                calculating R^2. Default is the training data.
            y - An optional DataFrame of the response data to be used for
                calculating R^2. Default is the training data.
            adjusted - A boolean indicating if the R^2 value is adjusted
                (True) or not (False).

        Returns:
            A real value of the computed R^2 value.
        """
        # Allow interfacing with sklearn's cross fold validation
        # self.fit(X, y)
        if X is None:
            X = self.training_data
        if y is None:
            y = self.y_train_

        pred = self.predict(X)
        sse = ((y - pred.iloc[:, 0]) ** 2).sum()
        ssto = ((y - y.mean()) ** 2).sum()

        if adjusted:
            numerator = sse
            denominator = ssto
        else:
            numerator = sse / (len(y) - len(self.X_train_.columns) - 2)
            denominator = ssto / (len(y) - 1)

        return 1 - numerator / denominator

    def score(self, X=None, y=None, adjusted=False, **kwargs):
        """Wrapper for sklearn api for cross fold validation.
        
        See LinearModel.r_squared.
        """
        return self.r_squared(X, y, adjusted, **kwargs)

    def _prediction_interval_width(self, X_new, alpha=0.05):
        """Helper function for calculating prediction interval widths."""
        mse = self.get_sse() / self.rdf
        s_yhat_squared = (X_new.dot(self.cov_) * X_new).sum(axis=1)
        s_pred_squared = mse + s_yhat_squared

        t_crit = stats.t.ppf(1 - (alpha / 2), self.rdf)

        return t_crit * (s_pred_squared ** 0.5)

    def _confidence_interval_width(self, X_new, alpha=0.05):
        """Helper function for calculating confidence interval widths."""
        _, p = X_new.shape
        s_yhat_squared = (X_new.dot(self.cov_) * X_new).sum(axis=1)
        # t_crit = stats.t.ppf(1 - (alpha / 2), n-p)
        W_crit_squared = p * stats.f.ppf(1 - (alpha / 2), p, self.rdf)
        return (W_crit_squared ** 0.5) * (s_yhat_squared ** 0.5)

    def plot(
        self,
        categorize_residuals=True,
        jitter=None,
        confidence_band=False,
        prediction_band=False,
        original_y_space=True,
        transformed_y_space=False,
        alpha=0.5,
        **kwargs
    ):
        """Visualizes the fitted LinearModel and its line of best fit.

        Arguments:
            categorize_residuals - A boolean that indicates if the residual
                points should be colored by categories (True) or not (False).
            jitter - A boolean that indicates if residuals should be jittered
                in factor plots (True) or not (False).
            confidence_band - A real value that specifies the width of the
                confidence band to be plotted. If band not desired, parameter
                is set to False.
            prediction_band - A real value that specifies the width of the
                prediction band to be plotted. If band not desired, parameter
                is set to False.
            y_space - A str that indicates the type of output space for the
                y-axis. If set to 't', the transformed space will be plotted.
                If set to 'o', the original or untransformed space will be
                plotted. If set to 'b', both will be plotted side-by-side.
            alpha - A real value that indicates the transparency of the
                residuals. Default is 0.5.
            kwargs - Additional named parameters that will be passed onto
                lower level matplotlib plotting functions.

        Returns:
            A matplotlib plot appropriate visualization of the model.
        """
        if confidence_band and prediction_band:
            raise Exception(
                "Only one of {confidence_band, prediction_band} may "
                "be set to True at a time."
            )

        terms = self.ex.reduce()

        if original_y_space and transformed_y_space:
            fig, (ax_o, ax_t) = plt.subplots(1, 2, **kwargs)
            y_spaces = ['o', 't']
            axs = [ax_o, ax_t]
        elif transformed_y_space:  # at least one of the two is False
            fig, ax_t = plt.subplots(1, 1, **kwargs)
            y_spaces = ['t']
            axs = [ax_t]
        elif original_y_space:
            fig, ax_o = plt.subplots(1, 1, **kwargs)
            y_spaces = ['o']
            axs = [ax_o]
        else:
            raise AssertionError(
                "At least one of either `original_y_space` or "
                "`transformed_y_space` should be True in "
                "`model.plot(...)` call."
            )

        for y_space_type, ax in zip(y_spaces, axs):
            original_y_space = y_space_type == "o"

            # Redundant, we untransform in later function calls
            # TODO: Fix later
            y_vals = self.y_train_
            if original_y_space:
                y_vals = self.re.untransform(y_vals)
            # Plotting Details:
            min_y = min(y_vals)
            max_y = max(y_vals)
            diff = (max_y - min_y) * 0.05
            min_y = min(min_y - diff, min_y + diff)  # Add a small buffer
            max_y = max(max_y - diff, max_y + diff)
            # TODO: Check if min() and max() are necessary here

            plot_args = {
                "categorize_residuals": categorize_residuals,
                "jitter": jitter,
                "terms": terms,
                "confidence_band": confidence_band,
                "prediction_band": prediction_band,
                "original_y_space": original_y_space,
                "alpha": alpha,
                "plot_objs": {
                    "figure": fig,
                    "ax": ax,
                    "y": {
                        "min": min_y,
                        "max": max_y,
                        "name": str(self.re)
                    }
                }
            }

            if len(terms['Q']) == 1:
                self._plot_one_quant(**plot_args)
            elif len(terms['Q']) == 0 and len(terms['C']) > 0:
                self._plot_zero_quant(**plot_args)  # TODO Make function
            else:
                raise Exception(
                    "Plotting line of best fit only expressions that reference"
                    " a single variable."
                )
        return fig

    def _plot_zero_quant(
        self,
        categorize_residuals,
        jitter,
        terms,
        confidence_band,
        prediction_band,
        original_y_space,
        alpha,
        plot_objs,
    ):
        """A helper function for plotting models in the case no quantitiative
        variables are present.
        """

        ax = plot_objs['ax']
        unique_cats = list(terms['C'])
        levels = [cat.levels for cat in unique_cats]
        level_amounts = [len(level_ls) for level_ls in levels]
        ml_index = level_amounts.index(max(level_amounts))
        ml_cat = unique_cats[ml_index]
        ml_levels = levels[ml_index]

        # List of categorical variables without the ml_cat
        cats_wo_most = unique_cats[:]
        cats_wo_most.remove(ml_cat)

        # List of levels for categorical variables without the ml_cat
        levels_wo_most = levels[:]
        levels_wo_most.remove(levels[ml_index])

        single_cat = len(cats_wo_most) == 0
        if single_cat:
            level_combinations = [None]
        else:
            level_combinations = product(*levels_wo_most)  # Cartesian product

        # To produce an index column to be used for the x-axis alignment
        line_x = pd.DataFrame({str(ml_cat): ml_levels}).reset_index()
        points = pd.merge(self.training_data, line_x, on=str(ml_cat))

        plot_objs['x'] = {'name': 'index'}

        points["<Y_RESIDS_TO_PLOT>"] = self.re.evaluate(points)
        if original_y_space:
            points["<Y_RESIDS_TO_PLOT>"] = self.re.untransform(
                points["<Y_RESIDS_TO_PLOT>"]
            )
            # Inefficient due to transforming, then untransforming.
            # Need to refactor later.

        plots = []
        labels = []
        linestyles = [':', '-.', '--', '-']
        for combination in level_combinations:
            points_indices = pd.Series([True] * len(points))
            if not single_cat:
                label = []
                for element, var in zip(combination, cats_wo_most):
                    name = str(var)
                    line_x[name] = element
                    label.append(str(element))

                    # Filter out points that don't apply to categories
                    points_indices = points_indices & (points[name] == element)
                labels.append(", ".join(label))
            line_type = linestyles.pop()
            linestyles.insert(0, line_type)
            line_y = self.predict(line_x, for_plot=True)
            y_vals = line_y["Predicted " + plot_objs['y']['name']]
            if original_y_space:
                y_vals_to_plot = self.re.untransform(y_vals)
            else:
                y_vals_to_plot = y_vals
            plot, = ax.plot(line_x.index, y_vals_to_plot, linestyle=line_type)
            if jitter is None or jitter is True:
                variability = np.random.normal(
                    scale=0.025,
                    size=sum(points_indices),
                )
            else:
                variability = 0
            # Y values must come from points because earlier merge
            # shuffles rows
            ax.scatter(
                points.loc[points_indices, 'index'] + variability,
                points.loc[points_indices, "<Y_RESIDS_TO_PLOT>"],
                c="black" if single_cat else plot.get_color(),
                alpha=alpha,
            )
            plots.append(plot)

            if confidence_band:
                self._plot_band(
                    line_x,
                    y_vals,
                    plot.get_color(),
                    original_y_space,
                    plot_objs,
                    True,
                    confidence_band,
                )
            elif prediction_band:
                self._plot_band(
                    line_x,
                    y_vals,
                    plot.get_color(),
                    original_y_space,
                    plot_objs,
                    False,
                    prediction_band,
                )

        if not single_cat and len(cats_wo_most) > 0:
            box = ax.get_position()
            ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
            ax.legend(
                plots,
                labels,
                title=", ".join([str(cat) for cat in cats_wo_most]),
                loc="center left",
                bbox_to_anchor=(1, 0.5),
            )
        ax.set_xlabel(str(ml_cat))
        ax.set_xticks(line_x.index)
        ax.set_xticklabels(line_x[str(ml_cat)])
        if not original_y_space:
            ax.set_ylabel(plot_objs['y']['name'])
        else:
            ax.set_ylabel(self.re.untransform_name())
        ax.grid()
        ax.set_ylim([plot_objs['y']['min'], plot_objs['y']['max']])

    def _plot_one_quant(
        self,
        categorize_residuals,
        jitter,
        terms,
        confidence_band,
        prediction_band,
        original_y_space,
        alpha,
        plot_objs,
    ):
        """A helper function for plotting models in the case only one
        quantitiative variable is present. Also support zero or more
        categorical variables."""
        # Get the "first" and only element in the set
        x_term = next(iter(terms['Q']))

        x_name = str(x_term)
        x = self.training_data[x_name]
        min_x = min(x)
        max_x = max(x)
        diff = (max_x - min_x) * 0.05
        min_x = min(min_x - diff, min_x + diff)  # Add a small buffer
        max_x = max(max_x - diff, max_x + diff)
        # TODO: Check if min() and max() are necessary here

        plot_objs['x'] = {"min" : min_x, "max" : max_x, "name" : x_name}

        # Quantitative inputs
        line_x = pd.DataFrame({x_name: np.linspace(min_x, max_x, 100)})

        if len(terms['C']) == 0:
            self._plot_one_quant_zero_cats(
                x,
                line_x,
                jitter,
                terms,
                confidence_band,
                prediction_band,
                original_y_space,
                alpha,
                plot_objs,
            )
        else:
            self._plot_one_quant_some_cats(
                x,
                line_x,
                categorize_residuals,
                jitter,
                terms,
                confidence_band,
                prediction_band,
                original_y_space,
                alpha,
                plot_objs,
            )

        ax = plot_objs['ax']
        ax.set_xlabel(x_name)
        if not original_y_space:
            ax.set_ylabel(plot_objs['y']['name'])
        else:
            ax.set_ylabel(self.re.untransform_name())
        ax.grid()
        ax.set_xlim([min_x, max_x])
        ax.set_ylim([plot_objs['y']['min'], plot_objs['y']['max']])

    def _plot_one_quant_zero_cats(
        self,
        x,
        line_x,
        jitter,
        terms,
        confidence_band,
        prediction_band,
        original_y_space,
        alpha,
        plot_objs,
    ):
        """A helper function for plotting models in the case only one
        quantitiative variable and no categorical variables are present."""

        x_name = plot_objs['x']['name']
        ax = plot_objs['ax']
        line_y = self.predict(line_x)
        y_vals = line_y["Predicted " + plot_objs['y']['name']]
        if original_y_space:
            y_vals_to_plot = self.re.untransform(y_vals)
        else:
            y_vals_to_plot = y_vals
        line_fit, = ax.plot(line_x[x_name], y_vals_to_plot)

        if confidence_band:
            self._plot_band(
                line_x,
                y_vals,
                line_fit.get_color(),
                original_y_space,
                plot_objs,
                True,
                confidence_band,
            )
        elif prediction_band:
            self._plot_band(
                line_x,
                y_vals,
                line_fit.get_color(),
                original_y_space,
                plot_objs,
                False,
                prediction_band,
            )

        y_train_vals = self.y_train_
        if original_y_space:
            y_train_vals = self.re.untransform(y_train_vals)

        ax.scatter(x, y_train_vals, c = "black", alpha = alpha)

    def _plot_band(
        self,
        line_x,
        y_vals,
        color,
        original_y_space,
        plot_objs,
        use_confidence=False,
        alpha=0.05,
    ):
        """A helper function to plot the confidence or prediction bands for
        a model. By default will plot prediction bands."""
        x_name = plot_objs['x']['name']
        X_new = self.ex.evaluate(line_x, fit = False)
        if self.intercept:
            n, _ = X_new.shape
            X_new = np.hstack((X_new, np.ones((n, 1))))

        if use_confidence:
            widths = self._confidence_interval_width(X_new, alpha)
        else:
            widths = self._prediction_interval_width(X_new, alpha)

        lower = y_vals - widths
        upper = y_vals + widths

        if original_y_space:
            lower = self.re.untransform(lower)
            upper = self.re.untransform(upper)

        plot_objs['ax'].fill_between(
            x=line_x[x_name],
            y1=lower,
            y2=upper,
            color=color,
            alpha=0.3,
        )


    def _plot_one_quant_some_cats(
        self,
        x,
        line_x,
        categorize_residuals,
        jitter,
        terms,
        confidence_band,
        prediction_band,
        original_y_space,
        alpha,
        plot_objs,
    ):
        """A helper function for plotting models in the case only one
        quantitiative variable and one or more categorical variables are
        present."""

        ax = plot_objs['ax']
        x_name = plot_objs['x']['name']
        y_name = plot_objs['y']['name']


        plots = []
        labels = []
        linestyles = [':', '-.', '--', '-']

        cats = list(terms['C'])
        cat_names = [str(cat) for cat in cats]
        levels = [cat.levels for cat in cats]
        # cartesian product of all combinations
        level_combinations = product(*levels)

        dummy_data = line_x.copy() # rest of columns set in next few lines

        y_train_vals = self.y_train_
        if original_y_space:
            y_train_vals = self.re.untransform(y_train_vals)

        for level_set in level_combinations:
            label = [] # To be used in legend
            for (cat,level) in zip(cats,level_set):
                dummy_data[str(cat)] = level # set dummy data for prediction
                label.append(str(level))

            line_type = linestyles.pop() # rotate through line styles
            linestyles.insert(0, line_type)

            line_y = self.predict(dummy_data, for_plot=True)
            y_vals = line_y["Predicted " + y_name]
            if original_y_space:
                y_vals_to_plot = self.re.untransform(y_vals)
            else:
                y_vals_to_plot = y_vals
            plot, = ax.plot(dummy_data[x_name], y_vals_to_plot, linestyle=line_type)
            plots.append(plot)
            labels.append(", ".join(label))


            if categorize_residuals:
                indices_to_use = pd.Series([True] * len(x)) # gradually gets filtered out
                for (cat,level) in zip(cats,level_set):
                    indices_to_use = indices_to_use & (self.training_data[str(cat)] == level)
                ax.scatter(
                    x[indices_to_use],
                    y_train_vals[indices_to_use],
                    c=plot.get_color(),
                    alpha=alpha,
                )

            if confidence_band:
                self._plot_band(
                    dummy_data,
                    y_vals,
                    plot.get_color(),
                    original_y_space,
                    plot_objs,
                    True,
                    confidence_band,
                )
            elif prediction_band:
                self._plot_band(
                    dummy_data,
                    y_vals,
                    plot.get_color(),
                    original_y_space,
                    plot_objs,
                    False,
                    prediction_band,
                )

        # Legend
        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
        ax.legend(
            plots,
            labels,
            title=", ".join(cat_names),
            loc="center left",
            bbox_to_anchor=(1, 0.5),
        )

        if not categorize_residuals:
            ax.scatter(x, y_train_vals, c="black", alpha=alpha)

    def residual_plots(self, **kwargs):
        """Plot the residual plots of the model.

        Arguments:
            kwargs - Named parameters that will be passed onto lower level
                matplotlib plotting functions.

        Returns:
            A tuple containing the matplotlib (figure, list of axes) for the
            residual plots.
        """
        terms = self.X_train_.columns
        fig, axs = plt.subplots(1, len(terms), **kwargs)
        for term, ax in zip(terms, axs):
            ax.scatter(self.X_train_.get_column(term), self.residuals_)
            ax.set_xlabel(str(term))
            ax.set_ylabel("Residuals")
            ax.set_title(str(term) + " v. Residuals")
            ax.grid()
        return fig, axs

    def partial_plots(self, alpha=0.5, **kwargs):
        """Plot the partial regression plots for the model

        Arguments:
            alpha - A real value indicating the transparency of the residuals.
                Default is 0.5.
            kwargs - Named parameters that will be passed onto lower level
                matplotlib plotting functions.

        Returns:
            A tuple containing the matplotlib (figure, list of axes) for the
            partial plots.
        """
        #terms = self.ex.flatten(separate_interactions = False)
        terms = self.ex.get_terms()
        fig, axs = plt.subplots(1, len(terms), **kwargs)

        for i, ax in zip(range(0, len(terms)), axs):

            xi = terms[i]

            sans_xi = Combination(terms[:i] + terms[i+1:])
            yaxis = LinearModel(sans_xi, self.re)
            xaxis = LinearModel(sans_xi, xi)

            yaxis.fit(self.training_data)
            xaxis.fit(self.training_data)

            ax.scatter(xaxis.residuals_, yaxis.residuals_, alpha=alpha)
            ax.set_title("Leverage Plot for " + str(xi))

        return fig, axs

    @staticmethod
    def ones_column(data):
        """Helper function to create a column of ones for the intercept."""
        return pd.DataFrame({"Intercept" : np.repeat(1, data.shape[0])})

    def plot_residual_diagnostics(self, **kwargs):
        """Produce a matrix of four diagnostic plots:
        the residual v. quantile plot, the residual v. fited values plot,
        the histogram of residuals, and the residual v. order plot.

        Arguments:
            kwargs - Named parameters that will be passed onto lower level
                matplotlib plotting functions.

        Returns:
            A tuple containing the matplotlib (figure, list of axes) for the
            partial plots.
        """

        f, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, **kwargs)
        self.residual_quantile_plot(ax=ax1)
        self.residual_fitted_plot(ax=ax2)
        self.residual_histogram(ax=ax3)
        self.residual_order_plot(ax=ax4)

        f.suptitle("Residal Diagnostic Plots for " + str(self))

        return f, (ax1, ax2, ax3, ax4)

    def residual_quantile_plot(self, ax=None):
        """Produces the residual v. quantile plot of the model.

        Arguments:
            ax - An optional parameter that is a pregenerated Axis object.

        Returns:
            A rendered matplotlib axis object.
        """
        if ax is None:
            _, ax = plt.subplots(1,1)

        stats.probplot(self.residuals_, dist="norm", plot=ax)
        ax.set_title("Residual Q-Q Plot")
        return ax

    def residual_fitted_plot(self, ax=None):
        """Produces the residual v. fitted values plot of the model.

        Arguments:
            ax - An optional parameter that is a pregenerated Axis object.

        Returns:
            A rendered matplotlib axis object.
        """
        if ax is None:
            _, ax = plt.subplots(1,1)

        ax.scatter(self.fitted_, self.residuals_)
        ax.set_title("Fitted Values v. Residuals")
        ax.set_xlabel("Fitted Value")
        ax.set_ylabel("Residual")

        return ax

    def residual_histogram(self, ax=None):
        """Produces the residual histogram of the model.

        Arguments:
            ax - An optional parameter that is a pregenerated Axis object.

        Returns:
            A rendered matplotlib axis object.
        """
        if ax is None:
            _, ax = plt.subplots(1,1)

        ax.hist(self.residuals_)
        ax.set_title("Histogram of Residuals")
        ax.set_xlabel("Residual")
        ax.set_ylabel("Frequency")

        return ax

    def residual_order_plot(self, ax=None):
        """Produces the residual v. order plot of the model.

        Arguments:
            ax - An optional parameter that is a pregenerated Axis object.

        Returns:
            A rendered matplotlib axis object.
        """
        if ax is None:
            _, ax = plt.subplots(1,1)

        ax.plot(self.training_data.index, self.residuals_, "o-")
        ax.set_title("Order v. Residuals")
        ax.set_xlabel("Row Index")
        ax.set_ylabel("Residual")

        return ax




import numpy as np
import math
"""Contains the logic for automatic model building (i.e. stepwise regression)."""

from abc import ABC, abstractmethod
from ordered_set import OrderedSet

from .model import LinearModel
from .comparison import _extract_dfs
from .expression import Constant


class Score(ABC):
    """Generic metric object for model evaluation."""

    def __init__(self, model, higher_is_better):
        self.higher_is_better = higher_is_better
        self.model = model

        if model is None:
            self._score = np.inf * (-1 if higher_is_better else 1)
        else:
            self._score = self.compute()

    @abstractmethod
    def compute(self):
        pass

    def __str__(self):
        return "{} | {}".format(type(self).__name__, self._score)

    def compare(self, other):
        """Return true if self is better than other based on 'higher_is_better'."""
        assert(type(self) is type(other))  
        # make sure we are not comparing different types of scores
        
        if self.higher_is_better:
            return self._score < other._score
        else:
            return self._score > other._score


class RSquared(Score):

    def __init__(self, model, adjusted=False):
        self.adjusted=adjusted

        super(RSquared, self).__init__(
            model=model,
            higher_is_better=True
        )

    def __str__(self):
        return "R^2 ({}adjusted) | {}".format(
            "" if self.adjusted else "un", 
            self._score,
        )

    def compute(self):
        """ Calculate the (adjusted) R^2 value of the model.

        Arguments:
            X - An optional DataFrame of the explanatory data to be used for
                calculating R^2. Default is the training data.
            Y - An optional DataFrame of the response data to be used for
                calculating R^2. Default is the training data.
            adjusted - A boolean indicating if the R^2 value is adjusted (True)
                or not (False).

        Returns:
            A real value of the computed R^2 value.
        """

        X = self.model.training_data
        y = self.model.training_y

        pred = self.model.predict(X)
        sse = ((y.iloc[:, 0] - pred.iloc[:, 0]) ** 2).sum()
        ssto = ((y.iloc[:, 0] - y.iloc[:, 0].mean()) ** 2).sum()

        if self.adjusted:
            numerator = sse
            denominator = ssto
        else:
            numerator = sse / (len(y) - len(self.model.X_train_.columns) - 2)
            denominator = ssto / (len(y) - 1)

        return 1 - numerator / denominator


class MSE(Score):

    def __init__(self, model):
        super(MSE, self).__init__(
            model=model,
            higher_is_better=False
        )

    def compute(self):
        dfs = _extract_dfs(self.model, dict_out=True)
        sse = self.model.get_sse()
        return sse / dfs["error_df"]


class MallowsCp(Score):

    def __init__(self, model):
        super(MallowsCp, self).__init__(
            model=model,
            higher_is_better=False,
        )

    def compute(self):
        dfs = _extract_dfs(self.model, dict_out=True)
        sse = self.model.get_sse()
        sigma_sq = self.model.std_err_est ** 2
        n, p = self.model.n, self.model.p

        return sse / sigma_sq - n + (2 * p)


class AIC(Score):

    def __init__(self, model):
        super(AIC, self).__init__(
            model=model,
            higher_is_better=False,
        )

    def compute(self):
        p = self.model.p
        log_likelihood = self.model.log_likelihood()

        return 2 * (p - log_likelihood)


class BIC(Score):

    def __init__(self, model):
        super(BIC, self).__init__(
            model=model,
            higher_is_better=False,
        )

    def compute(self):
        n, p = self.model.n, self.model.p
        log_likelihood = self.model.log_likelihood()

        return math.log(n) * p - 2 * log_likelihood

"""All metrics that are supported by default.""" 
_metrics = dict(
    r_squared=RSquared,
    r_squared_adjusted=lambda model: RSquared(model=model, adjusted=True),
    mse=MSE,
    cp=MallowsCp,
    aic=AIC,
    bic=BIC,
)


def stepwise(
    full_model,
    metric_name,
    forward=False,
    naive=False,
    data=None,
    verbose=False,
):
    """Perform forward or backward stepwise regression.
    
    Arguments:
        full_model - A model object that contains all of the terms to be 
            considered for the procedure.
        metric_name - A string containing the name of the metric to use in
            the procedure. Options include: "r_squared", "r_squared_adjusted",
            "mse", "cp", "aic", and "bic".
        forward - If True, specifies forwards stepwise regression. If False,
            specifies backwards stepwise regression. Default is False.
        naive - If True, allows for the removal or addition of terms in a
            model that depend on others being present (e.g. removing variable
            'X' while an interaction between 'X' and 'Y' are still present).
            Defaults to False.
        data - A dataframe that, if specified, will be used for the stepwise
            regression. If not specified, it is assumed the full_model has
            already been trained using some other data.
        verbose - If True, will print to console periodic updates.
            Default is False.

    Returns:
         A dictionary containing the metric used, the best value achieved, 
         and the associated best model (ranked via the chosen metric).
    """

    if data is not None:
        full_model.fit(data)

    metric_name = metric_name.lower()

    ex_terms = full_model.ex
    re_term = full_model.re
    data = full_model.training_data

    if ex_terms is None or re_term is None:
        raise AssertionError("The full model must be fit prior to undergoing a stepwise procedure.")

    if metric_name not in _metrics:
        raise KeyError("Metric '{}' not supported. The following metrics are supported: {}".format(
            metric_name,
            list(_metrics.keys())
        ))

    metric_func = _metrics[metric_name]

    ex_term_list = ex_terms.get_terms()
    if forward:
        best_model = LinearModel(Constant(1), re_term)
        best_model.fit(data)
    else:
        best_model = full_model

    best_metric = metric_func(best_model)
      

    while len(ex_term_list) > 0:
        best_potential_metric = metric_func(None)
        best_potential_model = None
        best_idx = None
    
        if forward and not naive:
            ex_term_list_expression = None
            for t in ex_term_list:
                if ex_term_list_expression is None:
                    ex_term_list_expression = t
                else:
                    ex_term_list_expression = ex_term_list_expression + t
            # Find all terms that do not depend on other terms
            leaves = OrderedSet(term for term in ex_term_list if not \
                term.contains(ex_term_list_expression - term))
    
        for i, term in enumerate(ex_term_list):
            try:
                if forward:
                    # validate if adding term is valid
                    if not naive:
                        if term not in leaves:
                            continue
                    potential_model = LinearModel(
                        best_model.given_ex + term, 
                        re_term,
                    )
                else:
                    # validate if removing term is valid
                    if not naive:
                        if (best_model.given_ex - term).contains(term):
                            continue
                    potential_model = LinearModel(
                        best_model.given_ex - term,
                        re_term,
                    )

                potential_model.fit(data)
                potential_metric = metric_func(potential_model)

                if best_potential_metric.compare(potential_metric):
                    best_potential_metric = potential_metric
                    best_potential_model = potential_model
                    best_idx = i

                if verbose:
                    print(potential_model)
                    print(potential_metric)
                    print("Current best potential model" if best_idx == i else "Not current best potential")
                    print()

            except np.linalg.linalg.LinAlgError:
                continue

        if best_metric.compare(best_potential_metric):
            best_metric = best_potential_metric
            best_model = best_potential_model
            if verbose:
                print("!!! New model found. Now including", ex_term_list[best_idx])
                print()
            del ex_term_list[best_idx]
        else:
            if verbose:
                print("!!! No potential models better than prior. Exiting search.")
                print()
            break
    else:
        if verbose:
            print("!!! Exhausted all potential terms. None left to consider.")


    return dict(
        forward=forward,
        metric=best_metric,
        metric_name=metric_name,
        best_model=best_model
    )



import unittest
from .expression import *
from .model import *
import pandas as pd

def floatComparison(a, b, eps=0.0001):
    if isinstance(a, (pd.Series, pd.DataFrame)) or isinstance(b, (pd.Series, pd.DataFrame)):
        return (a - eps < b) & (b < a + eps)
    else:
        return (a - eps < b) and (b < a + eps)

# Expressions
class TestVarMethods(unittest.TestCase):
    
    def test_str(self):
        self.assertEqual(str(Var("Test Case")), "Test Case")
        
    def test_mul(self):
        self.assertEqual(str(Var("Var1") * Var("Var2")), "(Var1)(Var2)")
        self.assertEqual(str(Var("Test") * Poly("Single", 2)), "(Single)(Test)+(Single^2)(Test)")
        
    def test_imul(self):
        v = Var("Test")
        v *= Var("Other")
        self.assertEqual(str(v), "(Other)(Test)")
        v = Var("Test")
        v *= Poly("Single", 2)
        self.assertEqual(str(v), "(Single)(Test)+(Single^2)(Test)")
                
    def test_flatten(self):
        flattened = Var("Test").get_terms()
        self.assertEqual(len(flattened), 1)
        self.assertEqual(str(flattened[0]), "Test")
                
    def test_copy(self):
        orig = Var("A")
        copy = orig.copy()
        self.assertFalse(orig is copy)
        self.assertEqual(str(orig), str(copy))
                
    def test_interpret(self):
        data = pd.DataFrame({"A" : [1,2,3], "B" : ["hello", "test", "goodbye"]})
        var1 = Var("A")
        var2 = Var("B")
        self.assertTrue(isinstance(var1.interpret(data), Quantitative))
        self.assertTrue(isinstance(var2.interpret(data), Categorical))
                
class TestQuantitativeMethods(unittest.TestCase):
    
    def test_str(self):
        var = 4*(Quantitative("A")+3)**2
        self.assertEqual(str(var), "4*(6*A+9+A^2)")
        var = 4*Sin(Quantitative("A")+3)
        self.assertEqual(str(var), "4*sin(3+A)")
        
    def test_add(self):
        var1 = Quantitative("A")
        self.assertEqual(str(var1 + 3), "3+A")
        self.assertEqual(str(var1 + var1), "2*A")
        
    def test_iadd(self):
        var = Quantitative("A")
        var += 3
        self.assertEqual(str(var), "3+A")
        var += -3
        var += var
        self.assertEqual(str(var), "2*A")
                
    def test_mul(self):
        var = Quantitative("A")
        self.assertEqual(str(var * 3), "3*A")
        self.assertEqual(str(var * var), "A^2")
        self.assertEqual(str(var * Quantitative("B")), "(A)(B)")
                
    def test_imul(self):
        var1 = Quantitative("A")
        var1 *= 3
        self.assertEqual(str(var1), "3*A")
        var2 = Quantitative("A")
        var2 *= var2
        self.assertEqual(str(var2), "A^2")
        var3 = Quantitative("A")
        var3 *= Quantitative("B")
        self.assertEqual(str(var3), "(A)(B)")
        
    def test_pow(self):
        var = Quantitative("A")
        self.assertEqual(str(var ** 4), "A^4")
        with self.assertRaises(Exception):
            var ** "bad arg"
                
    def test_ipow(self):
        var = Quantitative("A")
        var **= 3
        self.assertEqual(str(var), "A^3")
        with self.assertRaises(Exception):
            var *= "bad arg"
                
    def test_transform(self):
        var = Quantitative("A")
        var = var.transform("sin")
        self.assertEqual(str(var), "sin(A)")
        with self.assertRaises(Exception):
            var = Quantitative("A")
            var = var ** 3
            var.transform("bad call")

    def test_copy(self):
        orig = Quantitative("A")
        copy = orig.copy()
        self.assertFalse(orig is copy)
        self.assertEqual(str(orig), str(copy))

    def test_interpret(self):
        var = Quantitative("A")
        self.assertEqual(var.interpret(None), var)

class TestCategoricalMethods(unittest.TestCase):
    
    def test_str(self):
        var = Categorical("A")
        self.assertEqual(str(var), "A")
                
    def test_copy(self):
        orig = Categorical("A")
        copy = orig.copy()
        self.assertFalse(orig is copy)
        self.assertEqual(str(orig), str(copy))

    def test_interpret(self):
        var = Categorical("A")
        self.assertEqual(var.interpret(None), var)
                
class TestInteractionMethods(unittest.TestCase):
    
    def test_str(self):
        inter = Interaction(terms=(Quantitative("A"), Quantitative("B")))
        self.assertEqual(str(inter), "(A)(B)")
        
    def test_pow(self):
        inter = Interaction(terms=(Quantitative("A"), Quantitative("B")))
        inter = inter ** 3
        self.assertEqual(str(inter), "(A^3)(B^3)")
        with self.assertRaises(Exception):
            inter ** "log"
        
    def test_flatten(self):
        inter = Interaction(terms=(Quantitative("A"), Quantitative("B")))
        ls1 = inter.get_terms()
        self.assertEqual(len(ls1), 1)
        self.assertEqual(ls1[0], inter)
        ls2 = list(inter.terms)
        self.assertEqual(len(ls2), 2)
        self.assertTrue((str(ls2[0])=="A" and str(ls2[1])=="B") or (str(ls2[0])=="B" and str(ls2[1])=="A"))
                
    def test_copy(self):
        orig = Interaction(terms=(Var("A"), Var("B")))
        copy = orig.copy()
        self.assertFalse(orig is copy)
        self.assertEqual(str(orig), str(copy))
        
    def test_interpret(self):
        old = Var("A") * Var("B")
        data = pd.DataFrame({"A" : [1], "B" : ["cat"]})
        old.interpret(data)
        for v in old.terms:
            if v.name == "B":
                self.assertTrue(isinstance(v, Categorical))
            else:
                self.assertTrue(isinstance(v, Quantitative))
                
class TestCombinationMethods(unittest.TestCase):
    
    def test_str(self):
        comb = Var("A") + Var("B")
        self.assertEqual(str(comb), "A+B")
        
    def test_mul(self):
        comb = Var("A") + Var("B")
        self.assertEqual(str(comb * Var("C")), "(A)(C)+(B)(C)")
        with self.assertRaises(Exception):
            comb * "bad arg"
        
    def test_imul(self):
        comb = Var("A") + Var("B")
        comb *= Var("C")
        self.assertEqual(str(comb), "(A)(C)+(B)(C)")
        with self.assertRaises(Exception):
            comb *= "bad arg"
            
    def test_pow(self):
        comb = Var("A") + Var("B")
        self.assertEqual(str(comb ** 2), "2*(A)(B)+A^2+B^2")
                
    def test_flatten(self):
        comb = Var("A") + Var("B")
        ls = list(comb.get_terms())
        self.assertEqual(len(ls), 2)
        self.assertTrue((str(ls[0])=="A" and str(ls[1])=="B") or (str(ls[0])=="B" and str(ls[1])=="A"))

                
    def test_copy(self):
        orig = Interaction(terms=(Var("A"), Var("B")))
        copy = orig.copy()
        self.assertFalse(orig is copy)
        self.assertEqual(str(orig), str(copy))
        
    def test_interpret(self):
        old = Var("A") + Var("B")
        data = pd.DataFrame({"A" : [1], "B" : ["cat"]})
        old.interpret(data)
        for v in old.get_terms():
            if v.name == "B":
                self.assertTrue(isinstance(v, Categorical))
            else:
                self.assertTrue(isinstance(v, Quantitative))
        
iris = pd.read_csv("https://raw.githubusercontent.com/uiuc-cse/data-fa14/gh-pages/data/iris.csv")


# Model        
class TestModelMethods(unittest.TestCase):
    
    def test_fit(self):
        levels = ["virginica", "setosa", "versicolor"]
        explanatory = Q("petal_width") + C("species", levels = levels)
        response = Q("sepal_width")
        model = LinearModel(explanatory, response)
        results = model.fit(iris)[["Coefficient", "SE", "t", "p"]].sort_index()
        expected = pd.DataFrame({"Coefficient" : [1.36214962108944, 0.795582615454372, 1.86172822073969, 0.35290783081806], 
                                 "SE" : [0.248101683560026, 0.120657012161229, 0.223217037772943, 0.10358416791633], 
                                 "t" : [5.49, 6.59, 8.34, 3.41], 
                                 "p" : [0, 0, 0, 0.0008]}, 
                                 index = ["Intercept", "petal_width", "species{setosa}", "species{versicolor}"]).sort_index()
        diff = results - expected
        diff["Coefficient"] = floatComparison(0, diff["Coefficient"], 0.000001)
        diff["SE"] = floatComparison(0, diff["SE"], 0.000001)
        diff["t"] = floatComparison(0, diff["t"], 0.01)
        diff["p"] = floatComparison(0, diff["p"], 0.0001)

        self.assertTrue(all(diff.apply(all, 1)))
        
    def test_fit_types(self):
        n, p = 1000, 100
        num_iters = 10
        for i in range(num_iters):
            np.random.seed(i)
            # prepare data
            X_np, y_np = np.random.randn(n, p), np.random.randn(n)
            X_df = pd.DataFrame(X_np, columns=["x{}".format(i) for i in range(1,p+1)])
            y_df = pd.Series(y_np, name="y")
            df = pd.concat([X_df, y_df], axis=1)
            x_exp = sum(Q("x{}".format(i)) for i in range(1,p+1))
            y_exp = Q("y")

            # fit models with intercepts
            np_model = LinearModel(X_np, y_np, intercept=True)
            np_res = np_model.fit()[["Coefficient", "SE", "t", "p"]].sort_index()
            df_model = LinearModel(X_df, y_df, intercept=True)
            df_res = df_model.fit()[["Coefficient", "SE", "t", "p"]].sort_index()
            exp_model = LinearModel(x_exp, y_exp, intercept=True)
            exp_res = exp_model.fit(df)[["Coefficient", "SE", "t", "p"]].sort_index()

            diff = np_res - df_res
            diff["Coefficient"] = floatComparison(0, diff["Coefficient"], 0.000001)
            diff["SE"] = floatComparison(0, diff["SE"], 0.000001)
            diff["t"] = floatComparison(0, diff["t"], 0.01)
            diff["p"] = floatComparison(0, diff["p"], 0.0001)

            self.assertTrue(all(diff.apply(all, 1)))

            diff = np_res - exp_res
            diff["Coefficient"] = floatComparison(0, diff["Coefficient"], 0.000001)
            diff["SE"] = floatComparison(0, diff["SE"], 0.000001)
            diff["t"] = floatComparison(0, diff["t"], 0.01)
            diff["p"] = floatComparison(0, diff["p"], 0.0001)

            self.assertTrue(all(diff.apply(all, 1)))  
            # We assume that if these previous two tests are good then we don't need to compare df_res and exp_res

            # fit models without intercepts
            np_model = LinearModel(X_np, y_np, intercept=False)
            np_res = np_model.fit()[["Coefficient", "SE", "t", "p"]].sort_index()
            df_model = LinearModel(X_df, y_df, intercept=False)
            df_res = df_model.fit()[["Coefficient", "SE", "t", "p"]].sort_index()
            exp_model = LinearModel(x_exp, y_exp, intercept=False)
            exp_res = exp_model.fit(df)[["Coefficient", "SE", "t", "p"]].sort_index()

            diff = np_res - df_res
            diff["Coefficient"] = floatComparison(0, diff["Coefficient"], 0.000001)
            diff["SE"] = floatComparison(0, diff["SE"], 0.000001)
            diff["t"] = floatComparison(0, diff["t"], 0.01)
            diff["p"] = floatComparison(0, diff["p"], 0.0001)

            self.assertTrue(all(diff.apply(all, 1)))

            diff = np_res - exp_res
            diff["Coefficient"] = floatComparison(0, diff["Coefficient"], 0.000001)
            diff["SE"] = floatComparison(0, diff["SE"], 0.000001)
            diff["t"] = floatComparison(0, diff["t"], 0.01)
            diff["p"] = floatComparison(0, diff["p"], 0.0001)

            self.assertTrue(all(diff.apply(all, 1)))
            # We assume that if these previous two tests are good then we don't need to compare df_res and exp_res


    def test_predict(self):
        levels = ["virginica", "setosa", "versicolor"]
        explanatory = Q("petal_width") + C("species", levels=levels)
        response = Q("sepal_width")
        model = LinearModel(explanatory, response)
        model.fit(iris)
        newData = pd.DataFrame({"petal_width" : [1,2], "species" : ["setosa", "virginica"]})
        pred = model.predict(newData)
        expected = pd.DataFrame({"Predicted" : [4.019461, 3.306224]})
        diff = floatComparison(0, pred - expected, 0.00001)
        self.assertTrue(all(diff))
        
    def test_plot(self):
        model = LinearModel(Q("petal_width") + Q("petal_length"), Q("sepal_length"))
        model.fit(iris)
        with self.assertRaises(Exception):
            model.plot()
                
    def test_residual_plots(self):
        model = LinearModel(Q("petal_width") + Q("petal_length"), Q("sepal_length"))
        model.fit(iris)
        plots = model.residual_plots()
        self.assertEqual(len(plots), 2)
                
    def test_ones_column(self):
        ones = LinearModel.ones_column(iris)
        self.assertEqual(len(ones), 150)
        self.assertTrue(all(ones["Intercept"] == 1))
        
    '''        
    def test_extract_columns(self):
        self.assertEqual()
        with self.assertRaises(Exception):
            pass
        with self.assertRaises(Exception):
            pass
        with self.assertRaises(Exception):
            pass
        with self.assertRaises(Exception):
            pass
            
    def test_transform(self):
        self.assertEqual()
        with self.assertRaises(Exception):
            pass
        with self.assertRaises(Exception):
            pass            
    '''
    
if __name__ == "__main__":
    unittest.main()
        



"""Describes the object-oriented symbolic algebra."""

import collections
import numpy as np
from functools import reduce
from itertools import product
from abc import ABC, abstractmethod
from scipy.special import binom
from ordered_set import OrderedSet

from . import transformation as _t 

_supported_encodings = ['one-hot']

# if set to True, has expression representation equivalent to __str__ representation
# useful for debugging
STR_AS_REPR = False


class LightDataFrame(np.ndarray):

    def __new__(cls, input_array, columns=None):
        # Input array is an already formed ndarray instance
        # We first cast to be our class type
        obj = np.asarray(input_array, dtype=np.float64).view(cls)
        # add the new attribute to the created instance
        obj.columns = columns
        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # see InfoArray.__array_finalize__ for comments
        if obj is None: return
        self.columns = getattr(obj, 'columns', None)
    
    def get_column(self, colname):
        for i, column in enumerate(self.columns):
            if column == colname: 
                return self[:, i]
        raise KeyError("Column %s not in LightDataFrame." % colname)

    @property
    def index(self):
        return range(self.shape[0])



# ABC is a parent object that allows for Abstract methods
class Expression(ABC):
    """The parent abstract class that all subsequent representations of
    coefficients stem from and Model objects utilize."""
    
    def __init__(self, scale = 1):
        """Create an expression. Cannot be done directly, only through
        inheritance.

        Arguments:
            scale - A real value, default is 1. This represents a constant
                scaling that is applied to the Expression.

        Returns:
            An Expression object.
        """

        self.scale = scale
        
    @abstractmethod
    def __str__(self):
        """Represent an Expression object as a String utilizing standard
        algebraic notation."""
        pass

    def __repr__(self):
        global STR_AS_REPR
        if STR_AS_REPR:
            return self.__str__()
        else:
            return object.__repr__(self)
        
    def __eq__(self, other):
        """Check if an Expression is equivalent to another object.

        Arguments:
            other - An object to compare self to.

        Returns:
            A boolean that is True if they are equivalent, False if otherwise.
        """

        if isinstance(other, Expression):
            return self.scale == other.scale
        return False
    
    def __sim__(self, other):
        """Check if an Expression is similar to another object. This is done
        for the purposes of algebraic simplification.

        Arguments:
            other - An object to compare self to.

        Returns:
            A boolean that is True if they are similar, False if otherwise.
        """

        if isinstance(other, Expression):
            self_copy = self.copy()
            self_copy.scale = 1
            other_copy = other.copy()
            other_copy.scale = 1
            return self_copy == other_copy
        return False
        
    def __hash__(self):
        """Hash an Expression for the purposes of storing Expressions in sets
        and dictionaries.

        Returns:
            A real value that represents the hash of the object.
        """
        return hash(str(self))

    @abstractmethod
    def copy(self):
        """Creates a deep copy of an expression"""
        pass    
        
    @abstractmethod
    def interpret(self, data):
        """Cast a Var object or any nested Var objects in a functional manner
        to be either a Quantitative or Categorical variable.

        Arguments:
            data - A DataFrame object that contains a column == self.name

        Returns:
            A similar object that is either Quantitative or Categorical object.
        """       
        pass

    def transform(self, transformation):
        """Apply a transformation to an Expression object.

        Arguments:
            transformation - Either a transformation object of a str
                representation of common Transformations.

        Returns:
            A transformed Expression in the form of a TransVar object.
        """
        if isinstance(transformation, str):
            if transformation in _t._default_transformations:
                transformation = _t._default_transformations[transformation](None)
            else:
                raise Exception("Transformation specified is not a default function.")
        elif not isinstance(transformation, _t.Transformation):
            raise Exception("A Transformation or String are needed when transforming a variable.")
            
        self_copy = self.copy()
        return TransVar(self_copy, transformation)
        
    def __add__(self, other):
        """Form a combination of two Expressions by the addition operator (+).

        Arguments:
            other - An Expresssion object.

        Returns:
            Another Expression object that models the two operands added
            together. In most cases this is with a Combination object.
            Other cases it may be a simplified object of the same types
            as the original operands (i.e. X + X = 2*X).
        """
        if other == 0:
            return self
        elif isinstance(other, (int, float)):
            ret_exp = self.copy()
            shift = Constant(other)
            return ret_exp + shift
        elif self.__sim__(other):
            ret_exp = self.copy()
            ret_exp.scale += other.scale
            if ret_exp.scale == 0:
                return Constant(scale=0)
            else:
                return ret_exp
        elif isinstance(other, Combination):
            return Combination((self,)) + other  #other.__add__(self)
        elif isinstance(other, (Var, TransVar, Constant, Interaction)):
            # single term expressions
            return Combination((self, other))
        else:
            raise Exception("Expressions do not support addition with the given arguments.")
        
    def __radd__(self, other):
        """See __add__. """
        if isinstance(other, Expression):
            return other.__add__(self)
        else:
            return self.__add__(other)
    
    def __sub__(self, other):
        """Subtract one Expression object from another. This is a special case
        of addition. See __add__. """
        return self + -1 * other
    
    def __mul__(self, other):
        """Multiply two Expression objects together. When possible,
        simplification will be done. This is called with the multiplication
        operator (*).

        Arguments:
            other - An Expression object

        Returns:
            An Expression modeling the two operands multiplied together.
            If the inputs were single terms, then the output will be an
            Interaction object. If at least one was a combination of terms,
            then the output will most likely be a Combination object consisting
            of interactions as distribution will occur.
        """
        if isinstance(other, float) or isinstance(other, int):
            ret_exp = self.copy()
            ret_exp.scale *= other
            return ret_exp
        elif isinstance(other, Constant):
            ret_exp = self.copy()
            ret_exp.scale *= other.scale
            return ret_exp
        elif isinstance(other, Interaction):
            return other.__mul__(self)
        elif isinstance(other, Combination):
            self_copy = self.copy()
            return Combination(terms=(self_copy*t for t in other.get_terms()))
        elif isinstance(other, Expression):
            if self.__sim__(other): # Consolidate
                return self.copy() ** 2
            else: # Explicit interaction
                return Interaction(terms=(self, other))
        else:
            raise Exception("Expressions do not support multiplication with the given arguments.")
            
    def __rmul__(self, other):
        """See __mul__."""
        if isinstance(other, (int, float)):
            self_copy = self.copy()
            self_copy.scale *= other
            return self_copy
        elif isinstance(other, Constant):
            self_copy = self.copy()
            self_copy.scale *= other.scale
            return self_copy
        else:
            return other.__mul__(self)

    def __truediv__(self, other):
        """Divide one expression by another. This is a special case of __mul__
        and __pow__. """
        return self * (other ** -1)

    def __rtruediv__(self, other):
        """See __truediv__. """
        return other * (self ** -1)
    
    def __pow__(self, other): 
        """Raise an Expression object to a constant power. 
        
        Arguments:
            other - A real value.
        
        Returns:
            An Expression object that represents the original object raised to
            a constant power. More than likely this will either be a PowerVar
            object if the original object was a single term, or a
            Combination / Interaction if the original was one of those.
        """
        # This particualr function assumes a single term variable as this
        # operation for Combinations and Interactions are overloaded
        if other == 1:
            return self
        elif other == 0:
            return 1
        elif isinstance(other, (int, float)): # Proper power
            return PowerVar(self, other)
        else:
            raise Exception("Can only raise to a constant power.")

    def __and__(self, other):
        """Returns an interaction with main effects between two Expression
        objects."""
        return self + other + (self * other)

    def __xor__(self, other):
        """Returns an interaction with main effects up to a specified power.
        
        If the base object is not a Combination, then the default __pow__ logic
        is used."""

        return self ** other  # see Combination's __xor__ for other logic
        
    def descale(self):
        """Remove scaling on an Expression. Done in a functional manner so
        original object is unaffected."""
        ret_exp = self.copy()
        ret_exp._descale()
        return ret_exp
        
    def _descale(self): 
        """Helper function for descaling Expressions."""
        # Overwrite on expression containers such as Interactions and Combinations
        self.scale = 1
        
    @abstractmethod
    def evaluate(self, data, fit = True):
        """Given data, apply the appropriate transformations, combinations,
        and interactions.

        Arguments:
            data - A DataFrame whose column names match the names of the base
                Variable objects.
            fit - A flag to reference when evaluating the data to know when to
                overwrite Categorical levels.
        
        Returns:
            An appropriate DataFrame consisting of specified columns for each
            term in the Expression.
        """
        # Transform the incoming data depending on what type of variable the
        # data is represented by
        pass
    
    def reduce(self):
        """Obtain the base Quantitative, Categorical, Constant, and Varable
        terms.
        
        Returns:
            A dictionary mapping the respective types of terms to a set
            collection of the unique terms present.
        """
        return self._reduce({
            "Q":OrderedSet(),
            "C":OrderedSet(),
            "V":OrderedSet(),
            "Constant": None,
        })
    
    @abstractmethod
    def _reduce(self, ret_dict):
        """A helper method for the reduce function to allow for recursion."""
        # Reduce an expression to a dictionary containing lists of unique
        # Quantitative and Categorical Variables
        pass
    
    @abstractmethod
    def get_terms(self):
        """Return a list of the top level terms."""
        pass
    
    @abstractmethod
    def get_dof(self):
        """Return the degrees of freedom for an expression (not including
        constants)."""
        pass

    def untransform(self, data):
        """Untransform data by inverting the applied operations to it for
        the purposes of plotting. 
        
        Returns:
            A DataFrame object with appropriately inverted columns. 
        """
        return data * (1 / self.scale)

    def untransform_name(self):
        """Get the untransformed name of the model after inversions have
        been applied."""
        return str(self * (1 / self.scale))
    
    @abstractmethod
    def contains(self, other):
        """Returns true if this object is contains the other."""
        pass

class Var(Expression):
    """A base object that represents a generic single term. If a model consists
    of Y ~ X1 + X2 + X3, X2 would be a single term, as would X1 and X3. 
    
    This is more general when compared to a Quantitative or Categorical object,
    as it does not impose any restrictions on the data it represents. Upon
    model fitting, a Var will check to see if the data it represents is
    better suited for being a Quantitative variable or a Categorical one. 
    """

    def __init__(self, name, scale = 1):
        """Creates a Var object.

        Arguments:
            name - A str that will later be used to access a column in specific
                DataFrames. This name should coincide with the data it is
                representing.
            scale - A real value that will multiplicatively scale the term.
        """
        super().__init__(scale = scale)
        self.name = name
        
    def __eq__(self, other):
        if isinstance(other, Var):
            if self.name == other.name:
                return super().__eq__(other)
        return False
    
    def __hash__(self):
        return hash((self.name, self.scale))
        
    def __str__(self):
        if self.scale != 1:
            return "{1}*{0}".format(self.name, self.scale) 
        else:
            return str(self.name)
  
    def copy(self):
        return Var(self.name, self.scale)
        
    def interpret(self, data):
        if isinstance(data, LightDataFrame):
            if ('float' in data.get_column(self.name).dtype.name or \
                    'int' in data.get_column(self.name).dtype.name):
                return Quantitative(self.name, self.scale)
            else:
                return Categorical(self.name)
        else:  # isinstance(data, pd.DataFrame)
            if ('float' in data[self.name].dtype.name or \
                    'int' in data[self.name].dtype.name):
                return Quantitative(self.name, self.scale)
            else:
                return Categorical(self.name)
        
    def evaluate(self, data, fit=True):
        raise NotImplementedError("Must call interpret prior to evaluating data for variables.")
        
    def _reduce(self, ret_dict):
        ret_dict["V"].add(self)
        return ret_dict
    
    def get_terms(self):
        return [self]
    
    def get_dof(self):
        return 1
    
    def contains(self, other):
        return False  # impossible for a simple variable to contain another

class TransVar(Expression):
    """Represents an Expression that has a Transformation object applied to it.

    In the general case, Transformations do not support distributing across
    terms. As such, regardless of what the original Expression
    pre-transformation was, the new TransVar object will be treated as a single
    term.
    """
    
    def __init__(self, var, transformation, scale = 1):
        """Creates a TransVar object.

        Arguments:
            var - An Expression object to be transformed.
            transformation - A Transformation object to be applied to var.
            scale - A real value that will multiplicatively scale the term.
        """
        self.scale = scale
        self.var = var
        self.transformation = transformation
        
    def __str__(self):
        base = self.transformation.compose(str(self.var))
        if self.scale == 1:
            return str(base)
        else:
            return "{}*{}".format(self.scale, base)
        
    def copy(self):
        return TransVar(
            self.var.copy(),
            self.transformation.copy(),
            self.scale,
        )
    
    def __eq__(self, other):
        if isinstance(other, TransVar):
            return self.var == other.var and \
                   self.transformation == other.transformation and \
                   self.scale == other.scale
                    
        return False
    
    def __hash__(self):
        return hash((self.var, self.scale, self.transformation))
    
    def __add__(self, other):
        if isinstance(other, TransVar) and self.var == other.var and \
                self.transformation == other.transformation:
            return TransVar(
                self.var.copy(),
                self.transformation,
                self.scale + other.scale,
            )
        else:
            return Combination((self, other))
        
    def interpret(self, data):
        self.var = self.var.interpret(data)
        return self
    
    def _descale(self):
        self.scale = 1
        self.var._descale()
        
    def evaluate(self, data, fit = True):
        base_data = self.var.evaluate(data, fit).sum(axis=1)
        transformed_data = self.scale * self.transformation.transform(
            values=base_data,
            training=fit,
        )
        return LightDataFrame(
            transformed_data[:, np.newaxis],
            columns=[str(self)],
        )
    
    def _reduce(self, ret_dict):
        return self.var._reduce(ret_dict)
    
    def get_terms(self):
        return [self]
    
    def get_dof(self):
        return 1

    def untransform(self, data):
        unscaled = data  * (1 / self.scale)
        inverted = self.transformation.invert(unscaled)
        return self.var.untransform(inverted)

    def untransform_name(self):
        return self.var.untransform_name()

    def contains(self, other):
        if isinstance(other, Combination):
            return any(self.contains(other_term) for other_term in other.terms)

        return self.var.__eq__(other) or self.var.contains(other)
    
class PowerVar(TransVar):
    """A special case of the TransVar as a Power transformation can easily be
    distributed across terms. As such, special consideration is taken into
    account when multiplying two PowerVar objects.
    """
    def __init__(self, var, power, scale = 1):
        """Creates a PowerVar object.

        Arguments:
            var - An Expression object to be raised to a power.
            power - A real value to raise var to.
            scale - A real value that will multiplicatively scale the term.
        """
        self.scale = scale * (var.scale ** power)
        self.var = var.copy()
        self.var.scale = 1
        self.transformation = _t.Power(power)
        self.power = power
        
    def __eq__(self, other):
        if isinstance(other, PowerVar):
            return self.var == other.var and \
                   self.power == other.power and \
                   self.scale == other.scale

        return False
    
    def copy(self):
        return PowerVar(self.var, self.power, self.scale)
    
    def __hash__(self):
        return hash((self.var, self.scale, self.transformation, self.power))
        
    def __add__(self, other):
        if self.__sim__(other):
            return PowerVar(
                self.var.copy(),
                self.power,
                self.scale + other.scale,
            )
        else:
            return Combination((self, other))
        
    def __mul__(self, other):
        if isinstance(other, PowerVar):
            if self.var.__sim__(other.var):
                base = self.var.copy()
                base.scale = 1
                scalar = self.scale * (self.var.scale ** self.power) * \
                    (other.var.scale ** other.power)
                power = self.power + other.power
                if power == 0:
                    return Constant(scalar)
                else:
                    return PowerVar(self.var.copy(), power=power, scale=scalar)
        elif isinstance(other, Expression):
            if self.var.__sim__(other):
                base = self.var.copy()
                scalar = self.scale * (self.var.scale ** self.power) * \
                    (other.scale)
                power = self.power + 1
                if power == 0:
                    return Constant(scalar)
                else:
                    return PowerVar(self.var.copy(), power=power, scale=scalar)
            
        return super().__mul__(other)
    
    def __rmul__(self, other):
        if isinstance(other, Expression):
            return other.__mul__(self)
        else:
            return self.__mul__(other)
    
    def __pow__(self, other):
        if isinstance(other, int):
            return PowerVar(self.var.copy(), self.power * other)
        return super().__pow__(other)
    
    def get_terms(self):
        return [self]
    
    def get_dof(self):
        return 1

    def contains(self, other):
        if isinstance(other, Combination):
            return any(self.contains(other_term) for other_term in other.terms)

        if isinstance(other, PowerVar):
            if self.var.__eq__(other.var):
                return self.power > other.power
            else:
                return self.var.contains(other)
        else:
            return self.var.__eq__(other) or self.var.contains(other)
    
class Quantitative(Var):
    """A base term that inherits from Var. This specifically models
    Quantitative terms."""
    
    def __init__(self, name, scale = 1):
        """Creates a Quantitative object.

        Arguments:
            name - A str that will later be used to access a column in specific
                DataFrames. This name should coincide with the data it is
                representing.
            scale - A real value that will multiplicatively scale the term.
        """
        super().__init__(name = name, scale = scale)
        self.name = name
    
    def copy(self):
        return Quantitative(self.name, self.scale)
        
    def interpret(self, data):
        return self
    
    def evaluate(self, data, fit = True):
        if isinstance(data, LightDataFrame):
            transformed_data = self.scale * data.get_column(self.name)
        else:
            transformed_data = self.scale * data[self.name].values

        return LightDataFrame(
            transformed_data[:, np.newaxis], 
            columns=[self.name],
        )
    
    def _reduce(self, ret_dict):
        ret_dict["Q"].add(self)
        return ret_dict
    
    def get_terms(self):
        return [self]
    
    def get_dof(self):
        return 1
        
class Constant(Expression):
    """Represents a standalone term for constant values."""
    
    def __init__(self, scale = 1):
        """Create a Constant object.

        Arguments:
            scale - A real value that IS the constant value.
        """
        self.scale = scale
        
    def __str__(self):
        return str(self.scale)
    
    def __eq__(self, other):
        if isinstance(other, Constant):
            return self.scale == other.scale
        return False
    
    def __sim__(self, other):
        return isinstance(other, Constant)
    
    def __hash__(self):
        return hash((self.scale))
    
    def copy(self):
        return Constant(self.scale)
    
    def interpret(self, data):
        return self
    
    def __pow__(self, other):
        return Constant(self.scale ** other)
    
    def __mul__(self, other):
        if isinstance(other, Expression):
            ret_exp = other.copy()
            ret_exp.scale *= self.scale
            return ret_exp
        elif isinstance(other, (int, float)):
            return Constant(self.scale * other)
        
    def __rmul__(self, other):
        if isinstance(other, Expression):
            return other.__mul__(self)
        else:
            return self.__mul__(other)
    
    def evaluate(self, data, fit = True):
        n = len(data)
        if self.scale == 0:
            return LightDataFrame(np.empty(shape=(n, 0)), columns=[])
        else:
            return LightDataFrame(
                np.full((n, 1), self.scale),
                columns=[str(self)],
            )
        
    def _reduce(self, ret_dict):
        ret_dict['Constant'] = self.scale
        return ret_dict
    
    def get_terms(self):
        return list()
    
    def get_dof(self):
        if self.scale == 0:
            return 0
        else:
            return 1

    def contains(self, other):
        return False
    
class Categorical(Var):
    """The other base term that stems from the Var class. Represents solely
    categorical data."""

    def __init__(self, name, encoding='one-hot', levels=None, baseline=None):
        """Creates a Categorical object.

        Arguments:
            name - A str that will later be used to access a column in specific
                DataFrames. This name should coincide with the data it is
                representing.
            encoding - A str that represnts the supported encoding scheme to
                use. Default is one-hot.
            levels - A list object that holds all values to be considered as
                different levels during encoding. Any left out will be treated
                similarly as the baseline. A value of None will have levels
                learned upon fitting. 
            baseline - A list of objects to be collectively treated as a baseline.
        """
        self.scale = 1
        self.name = name
        if encoding not in _supported_encodings:
            raise Exception("Method " + str(encoding) + " not supported for Categorical variables.")
        self.encoding = encoding
        self.levels = levels
        self.baseline = baseline
        
    def __str__(self):
        return str(self.name)
        
    def copy(self):
        return Categorical(
            self.name,
            self.encoding,
            None if self.levels is None else self.levels[:],
            self.baseline,
        )
                
    def interpret(self, data):
        return self
    
    #def transform(self, transformation):
    #    raise Exception("Categorical variables cannot be transformed.")
        
    def set_baseline(self, value):
        if isinstance(value, collections.abc.Iterable) and \
                not isinstance(value, str):
            self.baseline = value
        else:
            self.baseline = [value]
        
    def _set_levels(self, data, override_baseline=True):
        unique_values = data[self.name].unique()
        unique_values.sort()
        if self.levels is None:
            self.levels = unique_values[:]
            if self.baseline is None:
                self.set_baseline(unique_values[0])
        else:
            unique_values = OrderedSet(unique_values)
            diff = unique_values - OrderedSet(self.levels)
            if len(diff) == 0:
                self.set_baseline(self.levels[0])
            else:
                self.set_baseline(diff)
                self.levels = list(self.levels)
                for element in diff:
                    self.levels.append(element)
        
        
    def _one_hot_encode(self, data):
        # keep track of the levels, including which ones are not in 
        # the baseline
        mapping = {}
        keep = []
        columns = []
        for i, level in enumerate(self.levels):
            mapping[level] = i
            if level not in self.baseline:
                keep.append(i)
                columns.append("%s{%s}" % (self.name, level))
        
        # define mapping of levels to integer codes
        if isinstance(data, LightDataFrame):
            codes = data.get_column(self.name).map(mapping)
        else:
            codes = data[self.name].map(mapping)

        # create dummy matrix by taking the appropriate rows from an 
        # identity matrix
        dummy_mat = np.eye(len(self.levels))[:, keep].take(codes, axis=0)
        
        return LightDataFrame(dummy_mat, columns=columns)
        
    def evaluate(self, data, fit=True):
        if self.levels is None or self.baseline is None or fit:
            self._set_levels(data)
        
        if self.encoding == 'one-hot':
            return self._one_hot_encode(data)
        else:
            raise NotImplementedError()
        
    def _reduce(self, ret_dict):
        ret_dict["C"].add(self)
        return ret_dict
    
    def get_terms(self):
        return [self]
    
    def get_dof(self):
        return len(self.levels) - len(self.baseline)
        
class Interaction(Expression):
    """An Interaction models two or more terms multiplied together. An
    Interaction is treated as a single term. If multiple terms are the result
    of a multiplication, then the result will be a Combination of Interactions.
    """

    def __init__(self, terms, scale = 1):
        """Create an Interaction object.

        Arguments:
            terms - A collection of terms to be modeled as multiplied together.
            scale - A real value that will multiplicatively scale the term.
        """
        self.scale = scale

        if any(not isinstance(t, Expression) for t in terms):
            raise Exception("Interaction takes only Expressions for initialization.")
        
        self.terms = OrderedSet()
        for term in terms:
            self._add_term(term)
        
    def __eq__(self, other):
        if isinstance(other, Interaction):
            if len(self.terms.difference(other.terms)) == 0 and \
                    len(other.terms.difference(self.terms)) == 0:
                return super().__eq__(other)
        return False
    
    def __hash__(self):
        return hash((frozenset(self.terms), self.scale))
        
    def __str__(self):
        if STR_AS_REPR:
            base = "(" + ")(".join((str(term) for term in self.terms)) + ")" 
        else:
            base = "(" + ")(".join(sorted(str(term) for term in self.terms)) + ")" 
        if self.scale == 1:
            return str(base)
        else:
            return "{}*{}".format(self.scale, base)
        
    def _add_term(self, other_term):
        other_term = other_term.copy()
        if isinstance(other_term, PowerVar):
            base_term = other_term.var
        else:
            base_term = other_term
        similar_term = None
        for term in self.terms:
            if term.__sim__(base_term) or (isinstance(term, PowerVar) and \
                    term.var.__sim__(base_term)):
                similar_term = term
                break

        self.scale *= other_term.scale
        other_term.scale = 1
        if similar_term is not None:
            self.terms.remove(similar_term)
            self.terms.add(similar_term * other_term)
        else:
            self.terms.add(other_term)               
   
    def _add_terms(self, other_terms):
        for term in other_terms:
            self._add_term(term)


    def copy(self):
        return Interaction(OrderedSet(term.copy() for term in self.terms), self.scale)
        
    def interpret(self, data):
        self.terms = OrderedSet(term.interpret(data) for term in self.terms)
        return self
    
    def __mul__(self, other):
        if isinstance(other, Interaction):
            self_copy = self.copy()
            self_copy._add_terms(other.terms)
            return self_copy
        elif isinstance(other, Constant):
            self_copy = self.copy()
            self_copy.scale *= other.scale
            return self_copy
        elif isinstance(other, Combination):
            self_copy = self.copy()
            return Combination(terms=(self_copy*t for t in other.get_terms()))
        elif isinstance(other, Expression):
            self_copy = self.copy()
            self_copy._add_term(other)
            return self_copy
        else:
            return super().__mul__(other)
        
    def __pow__(self, other):
        if isinstance(other, (int, float)):
            ret_int = Interaction(())
            for term in self.terms:
                ret_int._add_term(term ** other)
            return ret_int
        return super().__pow__(other)
    
    def _descale(self):
        self.scale = 1
        for term in self.terms:
            term._descale()
            
    def evaluate(self, data, fit=True):
        trans_data_sets = [term.evaluate(data, fit) for term in self.terms]
        # rename columns in sets
        for data_set in trans_data_sets:
            data_set.columns = ["({})".format(col) for col in data_set.columns]
            
        base_set = trans_data_sets[0]
        for other_set in trans_data_sets[1:]:
            new_data = []
            new_columns = []
            for base_col in base_set.columns:
                for other_col in other_set.columns:
                    new_data.append(base_set.get_column(base_col) * \
                        other_set.get_column(other_col))
                    new_columns.append(base_col + other_col)
            
            base_set = LightDataFrame(
                np.column_stack(new_data),
                columns=new_columns,
            )
            
        return base_set
    
    def _reduce(self, ret_dict):
        for term in self.terms:
            ret_dict = term._reduce(ret_dict)
            
        return ret_dict
    
    def get_terms(self):
        return [self]
    
    def get_dof(self):
        return reduce(lambda x,y: x*y, (term.get_dof() for term in self.terms))

    def contains(self, other):
        if isinstance(other, Combination):
            return any(self.contains(other_term) for other_term in other.terms)

        self_terms = self.terms
        if isinstance(other, Interaction):
            other_terms = other.terms
        else:
            other_terms = [other]
        all_terms_found = True
        for o_t in  other_terms:
            single_term_found = False
            for s_t in self_terms:
                single_term_found = single_term_found or s_t.__eq__(o_t)
                single_term_found = single_term_found or s_t.contains(o_t)
            all_terms_found = all_terms_found and single_term_found
        return all_terms_found
    
class Combination(Expression):
    """A Combination models several single terms added together. """

    def __init__(self, terms, scale = 1):
        """Create a Combination object.

        Arguments:
            terms - A collection of Expressions to add together.
            scale - A real value that will multiplicatively scale the term.
        """
        self.scale = scale
        
        terms = [Constant(t) if isinstance(t, (int, float)) else t for t in terms]

        if any(not isinstance(t, Expression) for t in terms):
            raise Exception("Combination takes only Expressions for initialization.")
                        
        self.terms = OrderedSet()
        for term in terms:
            self._add_term(term)
                    
    def __eq__(self, other):
        if isinstance(other, Combination):
            if len(self.terms.difference(other.terms)) == 0 and \
                    len(other.terms.difference(self.terms)) == 0:
                return super().__eq__(other)
        return False

    def __hash__(self):
        return hash((frozenset(self.terms), self.scale))
                
    def __str__(self):
        if STR_AS_REPR:
            base = "+".join((str(term) for term in self.terms)) 
        else:
            base = "+".join(sorted(str(term) for term in self.terms)) 
        if self.scale == 1:
            return str(base)
        else:
            return "{}*({})".format(self.scale, base)

    def __xor__(self, other):
        if isinstance(other, int) and other >= 0:
            power = min(other, len(self.terms))
            new_terms = OrderedSet()
            base_terms = self.terms
            last_added_terms = [frozenset()]

            for _ in range(power):
                newly_added_terms = OrderedSet()
                for new_term in base_terms:
                    for last_added_term in last_added_terms:
                        if new_term not in last_added_term:
                            newly_added_terms.add(
                                last_added_term.union(frozenset([new_term])),
                            )

                new_terms = new_terms.union(newly_added_terms)
                last_added_terms = newly_added_terms

            processed_terms = OrderedSet()
            for term in new_terms:
                if len(term) == 1:
                    processed_terms.add(next(iter(term)))
                else:
                    processed_terms.add(reduce(lambda x,y: x * y, term))

            if len(new_terms) == 0:
                return Constant(1)
            else:
                return Combination(processed_terms)
        else:
            return self ** other
        
    def _add_term(self, other_term):
        if isinstance(other_term, (int, float)):
            other_term = Constant(other_term)
        
        similar_term = None
        for term in self.terms:
            if term.__sim__(other_term):
                similar_term = term
                break
                
        if similar_term is not None:
            self.terms.remove(similar_term)
            addition_result = similar_term + other_term
            if addition_result != Constant(0):
                self.terms.add(addition_result)
        else:
            self.terms.add(other_term)       
            
    def _add_terms(self, other_terms):
        for term in other_terms:
            self._add_term(term)
        
    def copy(self):
        return Combination(OrderedSet(term.copy() for term in self.terms), self.scale)
        
    def interpret(self, data):
        self.terms = [term.interpret(data) for term in self.terms]
        return self
    
    def __add__(self, other):
        if isinstance(other, Combination):
            ret_comb = self.copy()
            ret_comb._add_terms(other.terms)
            return ret_comb
        elif isinstance(other, Expression):
            ret_comb = self.copy()
            ret_comb._add_term(other)
            return ret_comb
        else:
            return super().__add__(other)
        
    def __radd__(self, other):
        if isinstance(other, Expression):
            return other.__add__(self)
        else:
            return self.__add__(other)

    def __mul__(self, other):
        if isinstance(other, Combination):
            ret_comb = Combination(())
            for other_term in other.terms:
                ret_comb._add_terms((self * other_term).terms)
            return ret_comb
        elif isinstance(other, Expression):
            ret_comb = Combination(())
            for term in self.terms:
                ret_comb._add_term(term * other)
            return ret_comb
        else:
            return super().__mul__(other)
            
    def __pow__(self, other):
        if isinstance(other, int) and other > 0:
            terms = [self.scale * term for term in self.terms]
            return MultinomialExpansion(terms, other)
        else:
            return super().__pow__(other)
        
    def _descale(self):
        self.scale = 1
        for term in self.terms:
            term._descale()
            
    def evaluate(self, data, fit = True):
        dataframes = []
        columns = []
        for term in self.terms:
            df = term.evaluate(data, fit)
            dataframes.append(df)
            columns.extend(df.columns)
        
        return LightDataFrame(np.hstack(dataframes), columns=columns)
    
    def _reduce(self, ret_dict):
        for term in self.terms:
            ret_dict = term._reduce(ret_dict)
            
        return ret_dict
    
    def get_terms(self):
        return self.terms
    
    def get_dof(self):
        return reduce(lambda x,y: x+y, (term.get_dof() for term in self.terms))

    def contains(self, other):
        self_terms = self.terms
        if isinstance(other, Combination):
            other_terms = other.terms
        else:
            other_terms = [other]
        for s_t in self_terms:
            for o_t in  other_terms:
                if s_t.__eq__(o_t) or s_t.contains(o_t):
                    return True
        return False
    
def MultinomialCoef(params):
    """Calculate the coefficients necessary when raising polynomials to a power.

    Arguments:
        params - A list of powers of individual terms.

    Returns:
        An integer that is the coefficient for that term.
    """
    if len(params) == 1:
        return 1
    return binom(sum(params), params[-1]) * MultinomialCoef(params[:-1]) 

def MultinomialExpansion(terms, power):
    """Raise a collection of single terms (polynomial) to a power.

    Arguments:
        terms - A collection of Expression objects representing a polynomial.
        power - An integer to raise the polynomial to.

    Returns:
        A expanded / distributed Combination representing the polynomial raised
        to the specified power.
    """
    combination_terms = []
    generators = [range(power+1) for _ in range(len(terms))]
    for powers in product(*generators): # Inefficient, optimize later
        if sum(powers) == power:
            interaction_terms = [int(MultinomialCoef(powers))]
            for term, term_power in zip(terms, powers):
                if term_power == 0:
                    continue
                interaction_terms.append(term ** term_power)
            combination_terms.append(reduce(lambda x,y: x * y, interaction_terms))
    return reduce(lambda x,y: x + y, combination_terms)
    
def Poly(var, power):
    """A quick way to create a standard polynomial from one base expression.

    Arguments:
        var - An Expression object. 
        power - An integer to raise var to the power of.

    Returns:
        A Combination object. If var was a single term, then the standard
        polynomial is returned. If var was a Combination, then a Combination
        of all Interactions up to order 'power' is returned due to distribution
        rules.
    """
    if isinstance(var, str):
        var = Q(var)
    
    if not isinstance(power, int) or power < 0:
        raise Exception("Must raise to a non-negative integer power.")
    elif power == 0:
        return Constant(1)
    else:
        terms = [var ** i for i in range(1, power+1)]
        return reduce(lambda x,y: x + y, terms)
           
        
# Transformations 
Log = lambda var: var.transform("log")
Log10 = lambda var: var.transform("log10")
Sin = lambda var: var.transform("sin")
Cos = lambda var: var.transform("cos")
Exp = lambda var: var.transform("exp")
Standardize = lambda var: var.transform("standardize")
Z = lambda var: var.transform("standardize")
Cen = lambda var: var.transform("center")
Center = lambda var: var.transform("center")
Identity = lambda var: var.transform("identity")

# Aliases
V = Var
Q = Quantitative
Quant = Quantitative
C = Categorical
Cat = Categorical
Nominal = Categorical
Nom = Categorical
N = Categorical



from .model import *
from scipy.stats import f
from ordered_set import OrderedSet

import numpy as np
import pandas as pd

def anova(model1, model2 = None):
    """Perform inference by comparing two models.
    
    User-facing function to execute an Analysis of Variance for one or two
    models. Should only model be given, then a general F-test will be executed
    on all of the coefficients. Should two models be given, then a partial
    F-test will be executed. Note that one model needs to be a subset of the
    other for this to properly evaluate.

    Arguments:
        model1 - A Model object that has been fit on some data
        model2 - A Model object that has been fit on some data

    Returns:
        A DataFrame that contains relevant statistics for the test performed
    """
    if model2 is None:
        return _anova_terms(model1)
    elif is_subset(model1, model2):
        return _anova_models(model1, model2)
    elif is_subset(model2, model1):
        return _anova_models(model2, model1)
    else:
        raise Exception("Parameters must either be one model or two models where one is a subset of the other.")
        
def is_subset(model1, model2):
    """Checks if model1 contains all the terms of model2. In other words,
    checks if model2 is a subset of model1.

    Arguments:
        model1 - A Model object that has been fit on some data.
        model2 - A Model object that has been fit on some data.

    Returns: 
        A boolean value that is True if model2 is a subset of model1,
        False if model2 is not a subset of model1.       
    """
    if not model1.given_re.__sim__(model2.given_re):
        # Models should both have the same response variable
        return False
    
    terms1 = OrderedSet(model1.ex.get_terms())
    terms2 = OrderedSet(model2.ex.get_terms())
    return terms2.issubset(terms1)

def _calc_stats(numer_ss, numer_df, denom_ss, denom_df):
    """Given the appropriate sum of squares for the numerator and the mean sum
    of squares for the denominator (with respective degrees of freedom) this
    will return the relevant statistics of an F-test.

    Arguments:
        numer_ss - Sum of squares for the numerator.
        numer_df - Degrees of freedom for the numerator.
        denom_ms - Mean sum of squares for the denominator.
        denom_df - Degrees of freedom for the denominator.

    Returns:
        A tuple of three values: Element 0 contains the mean sum of squares for
        the numerator, element 1 contains the F statistic calculated, and 
        element 2 contains the associated p-value for the generated
        F-statistic.
    """
    numer_ms = numer_ss / numer_df
    denom_ms = denom_ss / denom_df
    f_val = numer_ms / denom_ms
    p_val = f.sf(f_val, numer_df, denom_df)
    return f_val, p_val

def _process_term(orig_model, term):
    """Obtains needed sum of squared residuals of a model fitted without a
    specified term/coefficient.

    Arguments:
        orig_model - A fitted Model object.
        term - A Variable object to be left out of the original model when
            fitting.

    Returns:
        A real value indicated the sum of squared residuals.
    """
    new_model = LinearModel(orig_model.given_ex - term, orig_model.given_re)
    new_model.fit(orig_model.training_data)
    return new_model.get_sse(), new_model.get_ssr()

def _extract_dfs(model, dict_out=False):
    """Obtains the different degrees of freedom for a model in reference to an
    F-test.

    Arguments:
        model - A fitted Model object

    Returns:
        A tuple containing three elements: Element 0 contains the degrees of
        freedom for the explantory variables, element 1 contains the degrees
        of freedom for the residuals, and element 2 contains the total degrees
        of freedom for the model.
    """
    reg_df = model.ex.get_dof()
    total_df = len(model.X_train_) - 1
    error_df = total_df - reg_df

    if dict_out:
        return dict(
            model_df=reg_df,
            total_df=total_df,
            error_df=error_df
        )
    else:
        return reg_df, error_df, total_df

def _anova_terms(model):
    """Perform a global F-test by analyzing all possible models when you leave
    one coefficient out while fitting.

    Arguments:
        model - A fitted model object.

    Returns:
        A DataFrame object that contains the degrees of freedom, adjusted sum
        of squares, adjusted mean sum of squares, F values, and p values for
        the associated tests performed.
    """
    full_reg_df, full_error_df, total_df = _extract_dfs(model)
  
    # Full model values
    full_sse = model.get_sse()  # sum of squared errors
    full_ssr = model.get_ssr()  # sum of squares explained by model
    full_sst = model.get_sst()  
    
    global_f_val, global_p_val = _calc_stats(
        full_ssr,
        full_reg_df,
        full_sse,
        full_error_df,
    )
    
    # Calculate the general terms now
    indices = ["Global Test"]
    sses = [full_ssr]
    ssrs = [full_ssr]
    f_vals = [global_f_val]
    p_vals = [global_p_val]
    dfs = [full_reg_df]
    
    terms = model.ex.get_terms()
    for term in terms:
        term_df = term.get_dof()
        reduced_sse, reduced_ssr = _process_term(model, term)
        reduced_f_val, reduced_p_val = _calc_stats(
            full_ssr - reduced_ssr,
            term_df,
            full_sse,
            full_error_df,
        )
        indices.append("- " + str(term))
        sses.append(reduced_sse)
        ssrs.append(reduced_ssr)
        dfs.append(term_df)
        f_vals.append(reduced_f_val)
        p_vals.append(reduced_p_val)
        
    # Finish off the dataframe's values
    indices.append("Error")
    sses.append("")
    ssrs.append("")
    dfs.append(full_error_df)
    f_vals.append("")
    p_vals.append("")
    
    return pd.DataFrame({
            "DF" : dfs,
            "SS Err.": sses,
            "SS Reg." : ssrs,
            "F" : f_vals,
            "p" : p_vals
        }, index = indices, columns = ["DF", "SS Err.", "SS Reg.", "F", "p"])

def _anova_models(full_model, reduced_model):
    """Performs a partial F-test to compare two models.

    Arguments:
        full_model - A fitted Model object.
        reduced_model - A fitted Model object that is a subset of the
            full_model.

    Returns:
        A DataFrame object that contains the resdiuals' degrees of freedom, sum
        of squares of the regression, degrees of freedom for the model, sum of
        squared residuals, the F value, and the p value for the associated test
        performed.
    """
    full_label = str(full_model)
    reduced_label = str(reduced_model)
    
    f_reg_df, f_error_df, f_total_df = _extract_dfs(full_model)
    r_reg_df, r_error_df, r_total_df = _extract_dfs(reduced_model)
    
    f_sse, f_ssr = full_model.get_sse(), full_model.get_ssr()
    r_sse, r_ssr = reduced_model.get_sse(), reduced_model.get_ssr()
  
    f_val, p_val = _calc_stats(
        r_sse - f_sse,
        r_error_df - f_error_df,
        f_sse,
        f_error_df,
    )
    
    indices = ["Full Model", "- Reduced Model", "Error"]
    df = [f_reg_df, f_reg_df - r_reg_df, f_error_df]
    ssrs = [f_ssr, r_ssr, ""] 
    sses = [f_sse, r_sse, ""]
    f = ["", f_val, ""]
    p = ["", p_val, ""]

    return pd.DataFrame({
        "DF" : df,
        "SS Err." : sses,
        "SS Reg." : ssrs, 
        "F" : f,
        "p" : p},
        index = indices, columns = ["DF", "SS Err.", "SS Reg.", "F", "p"])
    
    
    



from setuptools import setup, find_packages

setup(
    name="salmon-lm",
    version="1.2",

    description="A symbolic algebra based linear regression tool.",

    url="https://github.com/ajboyd2/salmon",
    download_url="https://github.com/ajboyd2/salmon/archive/v_100.tar.gz",

    author="Alex Boyd",
    author_email="alexjb@uci.edu",

    license="GPLv3",

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Education',
        'Topic :: Scientific/Engineering :: Mathematics',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
    ],
    
    keywords='modeling symbolic regression',
    install_requires=[
        "pandas>=1.5.3",
        "numpy>=1.25.0",
        "matplotlib>=3.7.1",
        "scipy>=1.10.1",
        "ordered-set>=4.1.0",
    ],
    packages=find_packages(exclude=("test",)),
)
